# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/NeuralActivityCubic',
                'doc_host': 'https://indoc-research.github.io',
                'git_url': 'https://github.com/Indoc-Research/NeuralActivityCubic',
                'lib_path': 'neuralactivitycubic'},
  'syms': { 'neuralactivitycubic.analysis': { 'neuralactivitycubic.analysis.AnalysisROI': ( 'analysis.html#analysisroi',
                                                                                            'neuralactivitycubic/analysis.py'),
                                              'neuralactivitycubic.analysis.AnalysisROI.__init__': ( 'analysis.html#analysisroi.__init__',
                                                                                                     'neuralactivitycubic/analysis.py'),
                                              'neuralactivitycubic.analysis.AnalysisROI._add_information_about_neighboring_intersections_to_peaks': ( 'analysis.html#analysisroi._add_information_about_neighboring_intersections_to_peaks',
                                                                                                                                                      'neuralactivitycubic/analysis.py'),
                                              'neuralactivitycubic.analysis.AnalysisROI._classify_area_under_curve_types': ( 'analysis.html#analysisroi._classify_area_under_curve_types',
                                                                                                                             'neuralactivitycubic/analysis.py'),
                                              'neuralactivitycubic.analysis.AnalysisROI._convert_polygon_to_mask': ( 'analysis.html#analysisroi._convert_polygon_to_mask',
                                                                                                                     'neuralactivitycubic/analysis.py'),
                                              'neuralactivitycubic.analysis.AnalysisROI._get_unique_frame_idxs_of_intersections_between_signal_and_baseline': ( 'analysis.html#analysisroi._get_unique_frame_idxs_of_intersections_between_signal_and_baseline',
                                                                                                                                                                'neuralactivitycubic/analysis.py'),
                                              'neuralactivitycubic.analysis.AnalysisROI._improve_intersection_frame_idx_estimation_by_interpolation': ( 'analysis.html#analysisroi._improve_intersection_frame_idx_estimation_by_interpolation',
                                                                                                                                                        'neuralactivitycubic/analysis.py'),
                                              'neuralactivitycubic.analysis.AnalysisROI.compute_amplitude_and_delta_f_over_f': ( 'analysis.html#analysisroi.compute_amplitude_and_delta_f_over_f',
                                                                                                                                 'neuralactivitycubic/analysis.py'),
                                              'neuralactivitycubic.analysis.AnalysisROI.compute_area_under_curve': ( 'analysis.html#analysisroi.compute_area_under_curve',
                                                                                                                     'neuralactivitycubic/analysis.py'),
                                              'neuralactivitycubic.analysis.AnalysisROI.compute_mean_intensity_timeseries': ( 'analysis.html#analysisroi.compute_mean_intensity_timeseries',
                                                                                                                              'neuralactivitycubic/analysis.py'),
                                              'neuralactivitycubic.analysis.AnalysisROI.compute_variance_area': ( 'analysis.html#analysisroi.compute_variance_area',
                                                                                                                  'neuralactivitycubic/analysis.py'),
                                              'neuralactivitycubic.analysis.AnalysisROI.detect_peaks': ( 'analysis.html#analysisroi.detect_peaks',
                                                                                                         'neuralactivitycubic/analysis.py'),
                                              'neuralactivitycubic.analysis.AnalysisROI.estimate_baseline': ( 'analysis.html#analysisroi.estimate_baseline',
                                                                                                              'neuralactivitycubic/analysis.py'),
                                              'neuralactivitycubic.analysis.BaselineEstimatorFactory': ( 'analysis.html#baselineestimatorfactory',
                                                                                                         'neuralactivitycubic/analysis.py'),
                                              'neuralactivitycubic.analysis.BaselineEstimatorFactory.get_baseline_estimation_callable': ( 'analysis.html#baselineestimatorfactory.get_baseline_estimation_callable',
                                                                                                                                          'neuralactivitycubic/analysis.py'),
                                              'neuralactivitycubic.analysis.BaselineEstimatorFactory.supported_baseline_estimation_methods': ( 'analysis.html#baselineestimatorfactory.supported_baseline_estimation_methods',
                                                                                                                                               'neuralactivitycubic/analysis.py')},
            'neuralactivitycubic.api': {'neuralactivitycubic.api.run_analysis': ('api.html#run_analysis', 'neuralactivitycubic/api.py')},
            'neuralactivitycubic.controller': { 'neuralactivitycubic.controller.App': ( 'controller.html#app',
                                                                                        'neuralactivitycubic/controller.py'),
                                                'neuralactivitycubic.controller.App.__init__': ( 'controller.html#app.__init__',
                                                                                                 'neuralactivitycubic/controller.py'),
                                                'neuralactivitycubic.controller.App._bind_buttons_of_view_to_functions_of_model': ( 'controller.html#app._bind_buttons_of_view_to_functions_of_model',
                                                                                                                                    'neuralactivitycubic/controller.py'),
                                                'neuralactivitycubic.controller.App._display_preview_of_representative_job': ( 'controller.html#app._display_preview_of_representative_job',
                                                                                                                               'neuralactivitycubic/controller.py'),
                                                'neuralactivitycubic.controller.App._load_data_button_clicked': ( 'controller.html#app._load_data_button_clicked',
                                                                                                                  'neuralactivitycubic/controller.py'),
                                                'neuralactivitycubic.controller.App._preview_window_size_button_clicked': ( 'controller.html#app._preview_window_size_button_clicked',
                                                                                                                            'neuralactivitycubic/controller.py'),
                                                'neuralactivitycubic.controller.App._run_button_clicked': ( 'controller.html#app._run_button_clicked',
                                                                                                            'neuralactivitycubic/controller.py'),
                                                'neuralactivitycubic.controller.App._setup_interaction_between_model_and_view': ( 'controller.html#app._setup_interaction_between_model_and_view',
                                                                                                                                  'neuralactivitycubic/controller.py'),
                                                'neuralactivitycubic.controller.App.launch': ( 'controller.html#app.launch',
                                                                                               'neuralactivitycubic/controller.py'),
                                                'neuralactivitycubic.controller.open_gui': ( 'controller.html#open_gui',
                                                                                             'neuralactivitycubic/controller.py')},
            'neuralactivitycubic.datamodels': { 'neuralactivitycubic.datamodels.BaseDataClass': ( 'datamodels.html#basedataclass',
                                                                                                  'neuralactivitycubic/datamodels.py'),
                                                'neuralactivitycubic.datamodels.BaseDataClass.display_all_attributes': ( 'datamodels.html#basedataclass.display_all_attributes',
                                                                                                                         'neuralactivitycubic/datamodels.py'),
                                                'neuralactivitycubic.datamodels.BaseDataClass.from_dict': ( 'datamodels.html#basedataclass.from_dict',
                                                                                                            'neuralactivitycubic/datamodels.py'),
                                                'neuralactivitycubic.datamodels.BaseDataClass.to_dict': ( 'datamodels.html#basedataclass.to_dict',
                                                                                                          'neuralactivitycubic/datamodels.py'),
                                                'neuralactivitycubic.datamodels.BaseDataClass.validate': ( 'datamodels.html#basedataclass.validate',
                                                                                                           'neuralactivitycubic/datamodels.py'),
                                                'neuralactivitycubic.datamodels.BaselineEstimationMethod': ( 'datamodels.html#baselineestimationmethod',
                                                                                                             'neuralactivitycubic/datamodels.py'),
                                                'neuralactivitycubic.datamodels.BaselineEstimationMethod.__str__': ( 'datamodels.html#baselineestimationmethod.__str__',
                                                                                                                     'neuralactivitycubic/datamodels.py'),
                                                'neuralactivitycubic.datamodels.Config': ( 'datamodels.html#config',
                                                                                           'neuralactivitycubic/datamodels.py'),
                                                'neuralactivitycubic.datamodels.Config.__post_init__': ( 'datamodels.html#config.__post_init__',
                                                                                                         'neuralactivitycubic/datamodels.py'),
                                                'neuralactivitycubic.datamodels.Config.to_json': ( 'datamodels.html#config.to_json',
                                                                                                   'neuralactivitycubic/datamodels.py'),
                                                'neuralactivitycubic.datamodels.Peak': ( 'datamodels.html#peak',
                                                                                         'neuralactivitycubic/datamodels.py'),
                                                'neuralactivitycubic.datamodels.ROIMode': ( 'datamodels.html#roimode',
                                                                                            'neuralactivitycubic/datamodels.py'),
                                                'neuralactivitycubic.datamodels.ROIMode.__str__': ( 'datamodels.html#roimode.__str__',
                                                                                                    'neuralactivitycubic/datamodels.py')},
            'neuralactivitycubic.input': { 'neuralactivitycubic.input.AVILoader': ('input.html#aviloader', 'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.AVILoader._get_all_frames': ( 'input.html#aviloader._get_all_frames',
                                                                                                    'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.Data': ('input.html#data', 'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.Data.__init__': ( 'input.html#data.__init__',
                                                                                        'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.Data._parse_loaded_data': ( 'input.html#data._parse_loaded_data',
                                                                                                  'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.DataLoader': ( 'input.html#dataloader',
                                                                                     'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.DataLoader.__init__': ( 'input.html#dataloader.__init__',
                                                                                              'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.DataLoader.load_and_parse_file_content': ( 'input.html#dataloader.load_and_parse_file_content',
                                                                                                                 'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.DataLoaderFactory': ( 'input.html#dataloaderfactory',
                                                                                            'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.DataLoaderFactory._assert_validity_of_filepath': ( 'input.html#dataloaderfactory._assert_validity_of_filepath',
                                                                                                                         'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.DataLoaderFactory._get_loader_for_file_extension': ( 'input.html#dataloaderfactory._get_loader_for_file_extension',
                                                                                                                           'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.DataLoaderFactory.all_supported_extensions': ( 'input.html#dataloaderfactory.all_supported_extensions',
                                                                                                                     'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.DataLoaderFactory.get_loader': ( 'input.html#dataloaderfactory.get_loader',
                                                                                                       'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.DataLoaderFactory.supported_extensions_per_data_loader': ( 'input.html#dataloaderfactory.supported_extensions_per_data_loader',
                                                                                                                                 'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.FocusAreaPathRestrictions': ( 'input.html#focusareapathrestrictions',
                                                                                                    'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.FocusAreaPathRestrictions.supported_dir_names': ( 'input.html#focusareapathrestrictions.supported_dir_names',
                                                                                                                        'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.GridWrapperROILoader': ( 'input.html#gridwrapperroiloader',
                                                                                               'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.GridWrapperROILoader._add_to_configs_and_create_as_attribute': ( 'input.html#gridwrapperroiloader._add_to_configs_and_create_as_attribute',
                                                                                                                                       'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.GridWrapperROILoader._get_boundary_row_col_coords_single_square': ( 'input.html#gridwrapperroiloader._get_boundary_row_col_coords_single_square',
                                                                                                                                          'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.GridWrapperROILoader._get_cropping_indices_to_adjust_for_window_size': ( 'input.html#gridwrapperroiloader._get_cropping_indices_to_adjust_for_window_size',
                                                                                                                                               'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.GridWrapperROILoader._get_row_col_idxs_of_grid': ( 'input.html#gridwrapperroiloader._get_row_col_idxs_of_grid',
                                                                                                                         'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.GridWrapperROILoader._get_row_col_labels_for_rois_in_grid': ( 'input.html#gridwrapperroiloader._get_row_col_labels_for_rois_in_grid',
                                                                                                                                    'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.GridWrapperROILoader.load_and_parse_file_content': ( 'input.html#gridwrapperroiloader.load_and_parse_file_content',
                                                                                                                           'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.GridWrapperROILoader.set_configs_for_grid_creation': ( 'input.html#gridwrapperroiloader.set_configs_for_grid_creation',
                                                                                                                             'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.ImageJROILoader': ( 'input.html#imagejroiloader',
                                                                                          'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.ImageJROILoader._extract_boundary_row_col_coords_from_roi_set': ( 'input.html#imagejroiloader._extract_boundary_row_col_coords_from_roi_set',
                                                                                                                                        'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.ImageJROILoader._extract_boundary_row_col_coords_from_single_roi': ( 'input.html#imagejroiloader._extract_boundary_row_col_coords_from_single_roi',
                                                                                                                                           'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.ImageJROILoader._get_boundary_row_col_coords_for_all_rois_in_source_data': ( 'input.html#imagejroiloader._get_boundary_row_col_coords_for_all_rois_in_source_data',
                                                                                                                                                   'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.NWBROILoader': ( 'input.html#nwbroiloader',
                                                                                       'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.NWBROILoader._get_boundary_row_col_coords_for_all_rois_in_source_data': ( 'input.html#nwbroiloader._get_boundary_row_col_coords_for_all_rois_in_source_data',
                                                                                                                                                'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.NWBRecordingLoader': ( 'input.html#nwbrecordingloader',
                                                                                             'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.NWBRecordingLoader._get_all_frames': ( 'input.html#nwbrecordingloader._get_all_frames',
                                                                                                             'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.ROI': ('input.html#roi', 'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.ROI._convert_to_valid_polygon': ( 'input.html#roi._convert_to_valid_polygon',
                                                                                                        'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.ROI._parse_loaded_data': ( 'input.html#roi._parse_loaded_data',
                                                                                                 'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.ROI.add_label_id': ( 'input.html#roi.add_label_id',
                                                                                           'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.ROI.create_nwb_compliant_pixel_mask': ( 'input.html#roi.create_nwb_compliant_pixel_mask',
                                                                                                              'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.ROILoader': ('input.html#roiloader', 'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.ROILoader._add_first_boundary_point_also_add_end_to_close_roi': ( 'input.html#roiloader._add_first_boundary_point_also_add_end_to_close_roi',
                                                                                                                                        'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.ROILoader._get_boundary_row_col_coords_for_all_rois_in_source_data': ( 'input.html#roiloader._get_boundary_row_col_coords_for_all_rois_in_source_data',
                                                                                                                                             'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.ROILoader.load_and_parse_file_content': ( 'input.html#roiloader.load_and_parse_file_content',
                                                                                                                'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.ROILoaderFactory': ( 'input.html#roiloaderfactory',
                                                                                           'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.ROILoaderFactory.supported_extensions_per_data_loader': ( 'input.html#roiloaderfactory.supported_extensions_per_data_loader',
                                                                                                                                'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.RecLoaderROILoaderCombinator': ( 'input.html#recloaderroiloadercombinator',
                                                                                                       'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.RecLoaderROILoaderCombinator.__init__': ( 'input.html#recloaderroiloadercombinator.__init__',
                                                                                                                'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.RecLoaderROILoaderCombinator._get_all_roi_loaders': ( 'input.html#recloaderroiloadercombinator._get_all_roi_loaders',
                                                                                                                            'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.RecLoaderROILoaderCombinator._get_the_recording_loader': ( 'input.html#recloaderroiloadercombinator._get_the_recording_loader',
                                                                                                                                 'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.RecLoaderROILoaderCombinator.get_all_recording_and_roi_loader_combos': ( 'input.html#recloaderroiloadercombinator.get_all_recording_and_roi_loader_combos',
                                                                                                                                               'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.Recording': ('input.html#recording', 'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.Recording._clip_image_at_bit_values': ( 'input.html#recording._clip_image_at_bit_values',
                                                                                                              'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.Recording._compute_contrast_and_brightness_adjustment_factors': ( 'input.html#recording._compute_contrast_and_brightness_adjustment_factors',
                                                                                                                                        'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.Recording._create_brightness_and_contrast_enhanced_preview': ( 'input.html#recording._create_brightness_and_contrast_enhanced_preview',
                                                                                                                                     'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.Recording._estimate_bit_depth': ( 'input.html#recording._estimate_bit_depth',
                                                                                                        'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.Recording._parse_loaded_data': ( 'input.html#recording._parse_loaded_data',
                                                                                                       'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.RecordingLoader': ( 'input.html#recordingloader',
                                                                                          'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.RecordingLoader._check_if_color_channels_are_redundant': ( 'input.html#recordingloader._check_if_color_channels_are_redundant',
                                                                                                                                 'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.RecordingLoader._check_if_multiple_color_channels_exist': ( 'input.html#recordingloader._check_if_multiple_color_channels_exist',
                                                                                                                                  'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.RecordingLoader._convert_to_grayscale': ( 'input.html#recordingloader._convert_to_grayscale',
                                                                                                                'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.RecordingLoader._ensure_array_is_4d': ( 'input.html#recordingloader._ensure_array_is_4d',
                                                                                                              'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.RecordingLoader._get_all_frames': ( 'input.html#recordingloader._get_all_frames',
                                                                                                          'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.RecordingLoader._load_all_frames': ( 'input.html#recordingloader._load_all_frames',
                                                                                                           'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.RecordingLoader.load_and_parse_file_content': ( 'input.html#recordingloader.load_and_parse_file_content',
                                                                                                                      'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.RecordingLoaderFactory': ( 'input.html#recordingloaderfactory',
                                                                                                 'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.RecordingLoaderFactory.supported_extensions_per_data_loader': ( 'input.html#recordingloaderfactory.supported_extensions_per_data_loader',
                                                                                                                                      'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.TIFFLoader': ( 'input.html#tiffloader',
                                                                                     'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.TIFFLoader._get_all_frames': ( 'input.html#tiffloader._get_all_frames',
                                                                                                     'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.get_filepaths_with_supported_extension_in_dirpath': ( 'input.html#get_filepaths_with_supported_extension_in_dirpath',
                                                                                                                            'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.test_successful_recording_loading': ( 'input.html#test_successful_recording_loading',
                                                                                                            'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.test_successful_roi_loading': ( 'input.html#test_successful_roi_loading',
                                                                                                      'neuralactivitycubic/input.py'),
                                           'neuralactivitycubic.input.test_unsupported_file_extension': ( 'input.html#test_unsupported_file_extension',
                                                                                                          'neuralactivitycubic/input.py')},
            'neuralactivitycubic.model': { 'neuralactivitycubic.model.Logger': ('model.html#logger', 'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Logger.__init__': ( 'model.html#logger.__init__',
                                                                                          'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Logger.add_new_log': ( 'model.html#logger.add_new_log',
                                                                                             'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Logger.clear_logs': ( 'model.html#logger.clear_logs',
                                                                                            'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Logger.get_logs': ( 'model.html#logger.get_logs',
                                                                                          'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Logger.save_current_logs': ( 'model.html#logger.save_current_logs',
                                                                                                   'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model': ('model.html#model', 'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model.__init__': ( 'model.html#model.__init__',
                                                                                         'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model._create_analysis_jobs_for_single_rec': ( 'model.html#model._create_analysis_jobs_for_single_rec',
                                                                                                                     'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model._create_single_analysis_job': ( 'model.html#model._create_single_analysis_job',
                                                                                                            'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model._display_configs': ( 'model.html#model._display_configs',
                                                                                                 'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model._ensure_data_from_previous_jobs_was_removed': ( 'model.html#model._ensure_data_from_previous_jobs_was_removed',
                                                                                                                            'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model._get_all_roi_loaders': ( 'model.html#model._get_all_roi_loaders',
                                                                                                     'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model._get_all_subdir_paths_with_rec_file': ( 'model.html#model._get_all_subdir_paths_with_rec_file',
                                                                                                                    'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model._get_focus_area_dir_path': ( 'model.html#model._get_focus_area_dir_path',
                                                                                                         'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model._get_recording_loader': ( 'model.html#model._get_recording_loader',
                                                                                                      'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model._save_user_settings_as_json': ( 'model.html#model._save_user_settings_as_json',
                                                                                                            'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model.add_info_to_logs': ( 'model.html#model.add_info_to_logs',
                                                                                                 'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model.create_analysis_jobs': ( 'model.html#model.create_analysis_jobs',
                                                                                                     'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model.preview_window_size': ( 'model.html#model.preview_window_size',
                                                                                                    'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model.run_analysis': ( 'model.html#model.run_analysis',
                                                                                             'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model.setup_connection_to_display_results': ( 'model.html#model.setup_connection_to_display_results',
                                                                                                                    'neuralactivitycubic/model.py'),
                                           'neuralactivitycubic.model.Model.setup_connection_to_update_infos_in_view': ( 'model.html#model.setup_connection_to_update_infos_in_view',
                                                                                                                         'neuralactivitycubic/model.py')},
            'neuralactivitycubic.processing': { 'neuralactivitycubic.processing.AnalysisJob': ( 'processing.html#analysisjob',
                                                                                                'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob.__init__': ( 'processing.html#analysisjob.__init__',
                                                                                                         'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._add_na3_results': ( 'processing.html#analysisjob._add_na3_results',
                                                                                                                 'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._add_source_data': ( 'processing.html#analysisjob._add_source_data',
                                                                                                                 'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._create_all_analysis_rois': ( 'processing.html#analysisjob._create_all_analysis_rois',
                                                                                                                          'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._create_analysis_rois': ( 'processing.html#analysisjob._create_analysis_rois',
                                                                                                                      'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._create_and_add_label_ids_to_all_rois_from_file': ( 'processing.html#analysisjob._create_and_add_label_ids_to_all_rois_from_file',
                                                                                                                                                'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._create_and_save_csv_result_files': ( 'processing.html#analysisjob._create_and_save_csv_result_files',
                                                                                                                                  'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._create_and_save_individual_traces_pdf_result_file': ( 'processing.html#analysisjob._create_and_save_individual_traces_pdf_result_file',
                                                                                                                                                   'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._create_and_save_single_trace_results_as_csv': ( 'processing.html#analysisjob._create_and_save_single_trace_results_as_csv',
                                                                                                                                             'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._create_results_dir': ( 'processing.html#analysisjob._create_results_dir',
                                                                                                                    'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._create_variance_area_dataframe': ( 'processing.html#analysisjob._create_variance_area_dataframe',
                                                                                                                                'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._export_all_data_to_nwb': ( 'processing.html#analysisjob._export_all_data_to_nwb',
                                                                                                                        'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._filter_rois_by_focus_area': ( 'processing.html#analysisjob._filter_rois_by_focus_area',
                                                                                                                           'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._get_info_for_nwb_units_table_from_analysis_roi': ( 'processing.html#analysisjob._get_info_for_nwb_units_table_from_analysis_roi',
                                                                                                                                                'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._initialize_nwb_file': ( 'processing.html#analysisjob._initialize_nwb_file',
                                                                                                                     'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._initialize_single_trace_df': ( 'processing.html#analysisjob._initialize_single_trace_df',
                                                                                                                            'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._load_data_from_all_roi_loaders': ( 'processing.html#analysisjob._load_data_from_all_roi_loaders',
                                                                                                                                'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._save_single_trace_df': ( 'processing.html#analysisjob._save_single_trace_df',
                                                                                                                      'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob._write_nwb_file': ( 'processing.html#analysisjob._write_nwb_file',
                                                                                                                'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob.create_results': ( 'processing.html#analysisjob.create_results',
                                                                                                               'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob.load_data_into_memory': ( 'processing.html#analysisjob.load_data_into_memory',
                                                                                                                      'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob.preview_window_size': ( 'processing.html#analysisjob.preview_window_size',
                                                                                                                    'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.AnalysisJob.run_analysis': ( 'processing.html#analysisjob.run_analysis',
                                                                                                             'neuralactivitycubic/processing.py'),
                                                'neuralactivitycubic.processing.process_analysis_rois': ( 'processing.html#process_analysis_rois',
                                                                                                          'neuralactivitycubic/processing.py')},
            'neuralactivitycubic.results': { 'neuralactivitycubic.results._get_all_peak_counts': ( 'results.html#_get_all_peak_counts',
                                                                                                   'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results._get_fontsize_adjusted_to_slimmest_roi': ( 'results.html#_get_fontsize_adjusted_to_slimmest_roi',
                                                                                                                     'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results._get_max_label_id_length': ( 'results.html#_get_max_label_id_length',
                                                                                                       'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results._get_max_peak_count': ( 'results.html#_get_max_peak_count',
                                                                                                  'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results._get_min_width_among_all_rois': ( 'results.html#_get_min_width_among_all_rois',
                                                                                                            'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results._get_text_bounding_box_size_in_data_dimensions': ( 'results.html#_get_text_bounding_box_size_in_data_dimensions',
                                                                                                                             'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results._get_total_peak_count': ( 'results.html#_get_total_peak_count',
                                                                                                    'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results._iteratively_decrease_fontsize_to_fit_text_in_squares': ( 'results.html#_iteratively_decrease_fontsize_to_fit_text_in_squares',
                                                                                                                                    'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results._plot_grid': ( 'results.html#_plot_grid',
                                                                                         'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results._plot_peak_count_text_for_each_analysis_roi': ( 'results.html#_plot_peak_count_text_for_each_analysis_roi',
                                                                                                                          'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results._plot_text_at_roi_centroid_coordinates': ( 'results.html#_plot_text_at_roi_centroid_coordinates',
                                                                                                                     'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results._use_grid_row_col_labels_as_axis_labels': ( 'results.html#_use_grid_row_col_labels_as_axis_labels',
                                                                                                                      'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results.create_single_roi_amplitude_and_delta_f_over_f_results': ( 'results.html#create_single_roi_amplitude_and_delta_f_over_f_results',
                                                                                                                                     'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results.create_single_roi_auc_results': ( 'results.html#create_single_roi_auc_results',
                                                                                                            'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results.export_peak_results_df_from_analysis_roi': ( 'results.html#export_peak_results_df_from_analysis_roi',
                                                                                                                       'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results.plot_activity_overview': ( 'results.html#plot_activity_overview',
                                                                                                     'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results.plot_intensity_trace_with_identified_peaks_for_individual_roi': ( 'results.html#plot_intensity_trace_with_identified_peaks_for_individual_roi',
                                                                                                                                            'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results.plot_roi_boundaries': ( 'results.html#plot_roi_boundaries',
                                                                                                  'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results.plot_rois_with_label_id_overview': ( 'results.html#plot_rois_with_label_id_overview',
                                                                                                               'neuralactivitycubic/results.py'),
                                             'neuralactivitycubic.results.plot_window_size_preview': ( 'results.html#plot_window_size_preview',
                                                                                                       'neuralactivitycubic/results.py')},
            'neuralactivitycubic.view': { 'neuralactivitycubic.view.AnalysisSettingsPanel': ( 'view.html#analysissettingspanel',
                                                                                              'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.AnalysisSettingsPanel.AnalysisSettings': ( 'view.html#analysissettingspanel.analysissettings',
                                                                                                               'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.AnalysisSettingsPanel.__init__': ( 'view.html#analysissettingspanel.__init__',
                                                                                                       'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.AnalysisSettingsPanel._configure_octaves_changed': ( 'view.html#analysissettingspanel._configure_octaves_changed',
                                                                                                                         'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.AnalysisSettingsPanel._include_variance_config_changed': ( 'view.html#analysissettingspanel._include_variance_config_changed',
                                                                                                                               'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.AnalysisSettingsPanel._limit_analysis_to_interval_changed': ( 'view.html#analysissettingspanel._limit_analysis_to_interval_changed',
                                                                                                                                  'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.AnalysisSettingsPanel.enable_analysis_settings': ( 'view.html#analysissettingspanel.enable_analysis_settings',
                                                                                                                       'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.AnalysisSettingsPanel.widget': ( 'view.html#analysissettingspanel.widget',
                                                                                                     'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.MainScreen': ('view.html#mainscreen', 'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.MainScreen.__init__': ( 'view.html#mainscreen.__init__',
                                                                                            'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.MainScreen.output_screen': ( 'view.html#mainscreen.output_screen',
                                                                                                 'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.MainScreen.show_output_screen': ( 'view.html#mainscreen.show_output_screen',
                                                                                                      'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.MainScreen.show_welcome_screen': ( 'view.html#mainscreen.show_welcome_screen',
                                                                                                       'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.MainScreen.welcome_screen': ( 'view.html#mainscreen.welcome_screen',
                                                                                                  'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.NWBMetadataPanel': ( 'view.html#nwbmetadatapanel',
                                                                                         'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.NWBMetadataPanel.NWBMetadata': ( 'view.html#nwbmetadatapanel.nwbmetadata',
                                                                                                     'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.NWBMetadataPanel.__init__': ( 'view.html#nwbmetadatapanel.__init__',
                                                                                                  'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.NWBMetadataPanel.widget': ( 'view.html#nwbmetadatapanel.widget',
                                                                                                'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.SourceDataPanel': ( 'view.html#sourcedatapanel',
                                                                                        'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.SourceDataPanel.SourceDataSettings': ( 'view.html#sourcedatapanel.sourcedatasettings',
                                                                                                           'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.SourceDataPanel.__init__': ( 'view.html#sourcedatapanel.__init__',
                                                                                                 'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.SourceDataPanel._change_batch_mode_config': ( 'view.html#sourcedatapanel._change_batch_mode_config',
                                                                                                                  'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.SourceDataPanel._change_data_source_path_configs': ( 'view.html#sourcedatapanel._change_data_source_path_configs',
                                                                                                                         'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.SourceDataPanel._change_focus_area_config': ( 'view.html#sourcedatapanel._change_focus_area_config',
                                                                                                                  'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.SourceDataPanel._change_roi_mode_config': ( 'view.html#sourcedatapanel._change_roi_mode_config',
                                                                                                                'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.SourceDataPanel.widget': ( 'view.html#sourcedatapanel.widget',
                                                                                               'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.UserInfoPanel': ( 'view.html#userinfopanel',
                                                                                      'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.UserInfoPanel.__init__': ( 'view.html#userinfopanel.__init__',
                                                                                               'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.UserInfoPanel.add_new_logs': ( 'view.html#userinfopanel.add_new_logs',
                                                                                                   'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.UserInfoPanel.update_progress_bar': ( 'view.html#userinfopanel.update_progress_bar',
                                                                                                          'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.UserInfoPanel.widget': ( 'view.html#userinfopanel.widget',
                                                                                             'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.WidgetsInterface': ( 'view.html#widgetsinterface',
                                                                                         'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.WidgetsInterface.__init__': ( 'view.html#widgetsinterface.__init__',
                                                                                                  'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.WidgetsInterface._data_source_path_chosen': ( 'view.html#widgetsinterface._data_source_path_chosen',
                                                                                                                  'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.WidgetsInterface._enable_window_size_widgets': ( 'view.html#widgetsinterface._enable_window_size_widgets',
                                                                                                                     'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.WidgetsInterface.adjust_widgets_to_loaded_data': ( 'view.html#widgetsinterface.adjust_widgets_to_loaded_data',
                                                                                                                       'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.WidgetsInterface.data_source_path': ( 'view.html#widgetsinterface.data_source_path',
                                                                                                          'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.WidgetsInterface.enable_analysis': ( 'view.html#widgetsinterface.enable_analysis',
                                                                                                         'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.WidgetsInterface.export_nwb_metadata': ( 'view.html#widgetsinterface.export_nwb_metadata',
                                                                                                             'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.WidgetsInterface.export_user_settings': ( 'view.html#widgetsinterface.export_user_settings',
                                                                                                              'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.WidgetsInterface.grid_size': ( 'view.html#widgetsinterface.grid_size',
                                                                                                   'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.WidgetsInterface.update_infos': ( 'view.html#widgetsinterface.update_infos',
                                                                                                      'neuralactivitycubic/view.py'),
                                          'neuralactivitycubic.view.change_widget_state': ( 'view.html#change_widget_state',
                                                                                            'neuralactivitycubic/view.py')}}}



"""Defines the graphical user interface based on interactive widgets."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_view.ipynb.

# %% auto 0
__all__ = ['change_widget_state', 'UserInfoPanel', 'NWBMetadataPanel', 'SourceDataPanel', 'AnalysisSettingsPanel', 'MainScreen',
           'WidgetsInterface']

# %% ../nbs/01_view.ipynb 3
import ipywidgets as w
from ipyfilechooser import FileChooser
from datetime import datetime
from dataclasses import dataclass, fields
from typing import Any
from pathlib import Path

# %% ../nbs/01_view.ipynb 4
from .datamodels import Config

# %% ../nbs/01_view.ipynb 5
def change_widget_state(widget: w.Widget,
                        value: Any | None = None,
                        description: str | None = None,
                        disabled: bool | None = None,
                        visibility: str | None = None,
                        tooltip: str | None = None,
                        button_style: str | None = None) -> w.Widget:
    if value is not None:
        widget.value = value
    if description is not None:
        widget.description = description
    if disabled is not None:
        widget.disabled = disabled
    if visibility is not None:
        widget.layout.visibility = visibility
    if tooltip is not None:
        widget.tooltip = tooltip
    if button_style is not None:
        widget.button_style = button_style
    return widget

# %% ../nbs/01_view.ipynb 6
class UserInfoPanel:

    def __init__(self) -> None:
        # Initialize basic parameters and configs:
        self.user_timezone = datetime.now().astimezone().tzinfo
        self.logs_message_count = 0
        self.progress_in_percent = 0.0 # still required?

        self.vertical_spacer = w.HTML(value = '', layout = w.Layout(height = '15px'))
        self.latest_logs = w.Label(
            value = ' ... the latest logs message will be displayed here ... ',
            style = {'font_style': 'italic', 'text_color': 'gray', 'font_family': 'monospace', 'text_align': 'center'},
            layout = w.Layout(width = 'initial')
        )
        self.progress_bar = w.FloatProgress(
            description = 'Progress',
            style = {'description_width': 'initial'},
            layout = w.Layout(width = '20%')
        )
        self.detailed_logs_output = w.Output(
            layout = w.Layout(
                max_height = '200px',
                y_overflow='scroll'
            )
        )
        self.max_log_lines_in_output = int(self.detailed_logs_output.layout.max_height.replace('px', '')) / 15
        self.detailed_logs_accordion = w.Accordion(
            children = [self.detailed_logs_output],
            titles = ('Detailed logs', ),
            selected_index = None,
            layout = w.Layout(width = '95%')
        )


    @property
    def widget(self) -> w.VBox:
        info = w.HTML(
            value="<p style='font-size:16px; font-weight:bold; text-align:center;'>Info</p>",
            layout = w.Layout(width = '40px')
        )
        info_overview_box = w.HBox(
            [
                info,
                self.latest_logs,
                self.progress_bar
            ],
            layout = w.Layout(
                justify_content = 'space-between',
                align_items = 'center',
                width = '95%'
            )
        )
        widget = w.VBox(
            [
                info_overview_box,
                self.detailed_logs_accordion,
                self.vertical_spacer
            ],
            layout = w.Layout(align_items = 'center')
        )
        widget.add_class('box-info-panel')
        return widget


    def add_new_logs(self, message: str) -> None:
        current_time = datetime.now(self.user_timezone)
        # Update latest logs:
        time_prefix_latest_logs = current_time.strftime('%H-%M-%S')
        if len(message) > 125:
            self.latest_logs.value = f'{time_prefix_latest_logs}: {message[:85]}...(find full message in detailed logs)'
        else:
            self.latest_logs.value = f'{time_prefix_latest_logs}: {message}'
        # Update detailed logs:
        time_prefix_detailed_logs = current_time.strftime('%d-%m-%y %H:%M:%S.%f')
        self.logs_message_count += 1
        # Check if title has already been changed:
        if self.detailed_logs_accordion.get_title(0) == 'Detailed logs':
            # If not, check if max lines have been reached with this new message and adjust title accordingly:
            if self.logs_message_count >= self.max_log_lines_in_output:
                self.detailed_logs_accordion.set_title(0, 'Detailed logs - please scroll down to see latest logs')
        with self.detailed_logs_output:
            print(f'{time_prefix_detailed_logs}: {message}')


    def update_progress_bar(self, progress_in_percent: float) -> None:
        if progress_in_percent == 100.0:
            self.progress_bar.bar_style = 'success'
        else:
            self.progress_bar.bar_style = ''
        self.progress_bar.value = progress_in_percent

# %% ../nbs/01_view.ipynb 7
class NWBMetadataPanel:
    @dataclass
    class NWBMetadata:
        experiment_description: w.Text = None
        experimenter: w.Text = None
        lab: w.Text = None
        institution: w.Text = None
        notes: w.Textarea = None
        model_number: w.Text = None
        manufacturer: w.Text = None
        description: w.Textarea = None
        location: w.Text = None
        calcium_indicator: w.Text = None
        excitation_wavelength: w.BoundedFloatText = None
        emission_wavelength: w.BoundedFloatText = None
        imaging_description: w.Textarea = None

        @property
        def current_metadata(self) -> dict[str, Any]:
            return {
                'experiment_description': self.experiment_description.value,
                'experimenter': self.experimenter.value,
                'lab': self.lab.value,
                'institution': self.institution.value,
                'notes': self.notes.value,
                'model_number': self.model_number.value,
                'manufacturer': self.manufacturer.value,
                'description': self.description.value,
                'location': self.location.value,
                'calcium_indicator': self.calcium_indicator.value,
                'excitation_wavelength': self.excitation_wavelength.value,
                'emission_wavelength': self.emission_wavelength.value,
                'imaging_description': self.imaging_description.value,
            }

    def __init__(self) -> None:
        self.metadata = self.NWBMetadata()
        self.metadata.experiment_description = w.Text()
        self.metadata.experimenter = w.Text()
        self.metadata.lab = w.Text()
        self.metadata.institution = w.Text()
        self.metadata.notes = w.Textarea()
        self.metadata.model_number = w.Text()
        self.metadata.manufacturer = w.Text()
        self.metadata.description = w.Textarea()
        self.metadata.location = w.Text()
        self.metadata.calcium_indicator = w.Text()
        self.metadata.excitation_wavelength = w.BoundedFloatText(min=0.0, max=1400.0)
        self.metadata.emission_wavelength = w.BoundedFloatText()
        self.metadata.imaging_description = w.Textarea()
        self.general_section = w.VBox(
            [
                w.HTML(
                    value="<p style='font-size:16px; font-weight:bold; text-align:center;'>General</p>"
                ),
                w.HBox(
                    [
                        w.Label(
                            'Experiment description: ',
                            layout=w.Layout(width='150px')
                        ),
                        self.metadata.experiment_description,
                    ],
                ),
                w.HBox(
                    [
                        w.Label(
                            'Experimenter: ',
                            layout=w.Layout(width='150px')
                        ),
                        self.metadata.experimenter,
                    ]
                ),
                w.HBox(
                    [
                        w.Label(
                            'Lab: ',
                            layout=w.Layout(width='150px')
                        ),
                        self.metadata.lab,
                    ]
                ),
                w.HBox(
                    [
                        w.Label(
                            'Institution: ',
                            layout=w.Layout(width='150px')
                        ),
                        self.metadata.institution,
                    ]
                ),
                w.HBox(
                    [
                        w.Label(
                            'Additional notes: ',
                            layout=w.Layout(width='150px')
                        ),
                        self.metadata.notes,
                    ]
                ),
            ],
            layout=w.Layout(
                justify_content='flex-start',
                align_items='flex-start',
                width='100%'
            )
        )
        self.microscope_section = w.VBox(
            [
                w.HTML(
                    value="<p style='font-size:16px; font-weight:bold; text-align:center;'>Microscope</p>"
                ),
                w.HBox(
                    [
                        w.Label(
                            'Model number: ',
                            layout=w.Layout(width='150px')
                        ),
                        self.metadata.model_number,
                    ],
                ),
                w.HBox(
                    [
                        w.Label(
                            'Manufacturer: ',
                            layout=w.Layout(width='150px')
                        ),
                        self.metadata.manufacturer,
                    ]
                ),
                w.HBox(
                    [
                        w.Label(
                            'Description: ',
                            layout=w.Layout(width='150px')
                        ),
                        self.metadata.description,
                    ]
                ),
            ],
            layout=w.Layout(
                justify_content='flex-start',
                align_items='flex-start',
                width='100%'
            )
        )
        self.imaging_section = w.VBox(
            [
                w.HTML(
                    value="<p style='font-size:16px; font-weight:bold; text-align:center;'>Imaging</p>"
                ),
                w.HBox(
                    [
                        w.Label(
                            'Location: ',
                            layout=w.Layout(width='150px')
                        ),
                        self.metadata.location,
                    ],
                ),
                w.HBox(
                    [
                        w.Label(
                            'Calcium indicator: ',
                            layout=w.Layout(width='150px')
                        ),
                        self.metadata.calcium_indicator,
                    ]
                ),
                w.HBox(
                    [
                        w.Label(
                            'Excitation wavelength: ',
                            layout=w.Layout(width='150px')
                        ),
                        self.metadata.excitation_wavelength,
                    ]
                ),
                w.HBox(
                    [
                        w.Label(
                            'Emission wavelength: ',
                            layout=w.Layout(width='150px')
                        ),
                        self.metadata.emission_wavelength,
                    ]
                ),
                w.HBox(
                    [
                        w.Label(
                            'Additional notes: ',
                            layout=w.Layout(width='150px')
                        ),
                        self.metadata.imaging_description,
                    ]
                ),
            ],
            layout=w.Layout(
                justify_content='flex-start',
                align_items='flex-start',
                width='100%'
            )
        )

    @property
    def widget(self) -> w.VBox:
        metadata_sections = w.HBox(
            [
                self.general_section,
                self.microscope_section,
                self.imaging_section,
            ],
            layout=w.Layout(width='100%')
        )
        nwb_metadata_accordion = w.Accordion(
            children=[metadata_sections],
            titles=('NWB Metadata',),
            selected_index=None,
            layout=w.Layout(width='95%')
        )
        info = w.HTML(
            value="<p style='font-size:16px; font-weight:bold; text-align:center;'>NWB Metadata</p>",
            layout=w.Layout(width='95%')
        )
        widget = w.VBox(
            [
                info,
                nwb_metadata_accordion,
                w.HTML(value='', layout=w.Layout(height='15px'))
            ],
            layout=w.Layout(
                align_items='center',
                border_bottom='1px solid',
            )
        )
        widget.add_class('box-info-panel')
        return widget

# %% ../nbs/01_view.ipynb 8
class SourceDataPanel:

    @dataclass
    class SourceDataSettings:
        batch_mode: w.Checkbox = None
        focus_area_enabled: w.Checkbox = None
        roi_mode: w.Dropdown = None
        data_source_path: FileChooser = None

        @property
        def current_values(self) -> dict[str, Any]:
            return {
                'batch_mode': self.batch_mode.value,
                'focus_area_enabled': self.focus_area_enabled.value,
                'roi_mode': self.roi_mode.value,
                'data_source_path': self.data_source_path.value
            }

    
    def __init__(self) -> None:
        self.vertical_spacer = w.HTML(value = '', layout = w.Layout(height = '5px'))
        self.settings = self.SourceDataSettings()
        self.load_source_data_button = w.Button(
            description = 'Load Data',
            disabled = True,
            tooltip = 'Please first select which source data to load!',
            layout = w.Layout(width = '12%', height = '55px')
        )

    @property
    def widget(self) -> w.HBox:
        # Create and configure all elements:
        self.settings.batch_mode = w.Checkbox(
            description = 'Enable batch processing',
            value = False,
            indent = False
        )
        self.settings.focus_area_enabled = w.Checkbox(
            description = 'Limit analysis to focus area',
            value = False,
            indent = False
        )
        self.settings.roi_mode = w.Dropdown(
            options = [
                ('Use adjustable grid to create ROIs (congruent squares)', 'grid'),
                ('Load predefined ROIs from source data', 'file')
            ],
            description = 'ROIs:',
            value = 'grid',
            style = {'description_width': 'initial'},
            layout = {'width': 'initial'}
        )
        self.settings.data_source_path = FileChooser(
            title = 'Please select the recording file:',
            layout = w.Layout(width = '50%')
        )
        self.settings.data_source_path.rows = 8

        # Enable event handling:
        self.settings.batch_mode.observe(self._change_batch_mode_config)
        self.settings.focus_area_enabled.observe(self._change_focus_area_config)
        self.settings.roi_mode.observe(self._change_roi_mode_config)

        io_source_data_info = w.HTML(
            value="<p style='font-size:16px; font-weight:bold; text-align:center;'>General Settings</p>"
        )
        processing_modes_box = w.VBox(
            [
                self.settings.roi_mode,
                w.HBox(
                    [
                        self.settings.batch_mode,
                        self.settings.focus_area_enabled
                    ]
                )
            ],
            layout = {'width': '33%'}
        )
        # Arrange elements:
        source_data_settings_box = w.HBox(
            [
                processing_modes_box,
                self.settings.data_source_path,
                self.load_source_data_button
            ],
            layout = w.Layout(width = '100%', justify_content = 'space-between')
        )
        general_settings_box = w.VBox(
            [
                io_source_data_info,
                source_data_settings_box,
                self.vertical_spacer
            ],
            layout = w.Layout(width = '95%', align_items = 'center')
        )
        general_settings_box.add_class('box-general-settings')
        # Return the complete widget:
        return w.HBox([general_settings_box], layout = w.Layout(width = '100%', justify_content = 'center'))


    def _change_data_source_path_configs(self, show_dirs_only: bool, title: str, reset: bool) -> None:
        if reset:
            self.settings.data_source_path.reset()
            change_widget_state(self.load_source_data_button,
                                disabled = True,
                                tooltip = 'Please select which source data to load!')
        self.settings.data_source_path.show_only_dirs = show_dirs_only
        self.settings.data_source_path.title = title
        
    
    def _change_roi_mode_config(self, change) -> None:
        if change['name'] == 'value':
            if change['new'] == 'file':
                show_only_dirs = True
                if self.settings.batch_mode.value:
                    title = 'Please select the parent directory that contains subdirectories with the individual source data:'
                elif self.settings.focus_area_enabled.value:
                    title = 'Please select the directory that contains the recording, all ROI files, and a directory with focus area ROI(s):'
                else:
                    title = 'Please select the directory that contains the recording and all ROI files:'
            else:
                if self.settings.batch_mode.value:
                    title = 'Please select the parent directory that contains subdirectories with the individual source data:'
                    show_only_dirs = True
                elif self.settings.focus_area_enabled.value:
                    title = 'Please select the directory that contains the recording and a directory with focus area ROI(s):'
                    show_only_dirs = True
                else:
                    title = 'Please select the recording file:'
                    show_only_dirs = False
            reset = self.settings.data_source_path.show_only_dirs != show_only_dirs
            self._change_data_source_path_configs(show_only_dirs, title, reset)
           

    def _change_focus_area_config(self, change) -> None:
        if change['name'] == 'value':
            roi_mode = self.settings.roi_mode.value
            fake_change = {'name': 'value', 'new': roi_mode}
            self._change_roi_mode_config(fake_change)
           

    def _change_batch_mode_config(self, change) -> None:
        if change['name'] == 'value':
            roi_mode = self.settings.roi_mode.value
            fake_change = {'name': 'value', 'new': roi_mode}
            self._change_roi_mode_config(fake_change)

# %% ../nbs/01_view.ipynb 9
class AnalysisSettingsPanel:

    @dataclass
    class AnalysisSettings:
        grid_size: w.IntSlider = None
        signal_to_noise_ratio: w.BoundedFloatText = None
        noise_window_size: w.BoundedIntText = None
        mean_signal_threshold: w.BoundedFloatText = None
        min_peak_count: w.BoundedIntText = None
        baseline_estimation_method: w.Dropdown = None
        include_variance: w.Checkbox = None
        variance_window_size: w.BoundedIntText = None
        use_frame_range: w.Checkbox = None
        frame_interval_to_analyze: w.IntRangeSlider = None
        customize_octave_filtering: w.Checkbox = None
        min_octave_span: w.BoundedFloatText = None
        save_overview_png: w.Checkbox = None
        save_summary_results: w.Checkbox = None
        save_single_trace_results: w.Checkbox = None
        export_to_nwb: w.Checkbox = None

        def __iter__(self):
            for field in fields(self):
                yield getattr(self, field.name)

        @property
        def current_values(self) -> dict[str, Any]:
            return {
                'grid_size': self.grid_size.value,
                'signal_to_noise_ratio': self.signal_to_noise_ratio.value,
                'noise_window_size': self.noise_window_size.value,
                'mean_signal_threshold': self.mean_signal_threshold.value,
                'min_peak_count': self.min_peak_count.value,
                'baseline_estimation_method': self.baseline_estimation_method.value,
                'include_variance': self.include_variance.value,
                'variance_window_size': self.variance_window_size.value,
                'use_frame_range': self.use_frame_range.value,
                'start_frame_idx': self.frame_interval_to_analyze.value[0],
                'end_frame_idx': self.frame_interval_to_analyze.value[1],
                'customize_octave_filtering': self.customize_octave_filtering.value,
                'min_octave_span': self.min_octave_span.value,
                'save_overview_png': self.save_overview_png.value,
                'save_summary_results': self.save_summary_results.value,
                'save_single_trace_results': self.save_single_trace_results.value,
                'export_to_nwb': self.export_to_nwb.value
            }

    def __init__(self) -> None:
        self.settings = self.AnalysisSettings()
        self.vertical_spacer = w.HTML(value = '', layout = w.Layout(height = '5px'))
        self.run_analysis_button = w.Button(
            description='Run Analysis',
            disabled=True,
            tooltip='You have to load some data first, before you can run the analysis!',
            button_style='',
            icon='rocket',
            layout=w.Layout(width='90%')
        )
        self.preview_window_size_button = w.Button(
            description = 'Preview',
            disabled = True,
            tooltip = 'Preview grid size. Does not start analysis',
            layout = w.Layout(width = '20%')
        )

    @property
    def widget(self) -> w.VBox:
        width_percentage_core_widgets = '95%'
        description_width = '35px'

        # Create and configure all elements:
        analysis_settings_info = w.HTML(
            value="<p style='font-size:16px; font-weight:bold; text-align:center;'>Analysis Settings</p>",
            layout = w.Layout(width = '99%')
        )

        # Create widgets for user settings:
        self.settings.grid_size = w.IntSlider(
            description = 'Grid size:',
            value = 10,
            min = 1,
            max = 128,
            step = 1,
            disabled = True,
            layout = w.Layout(width = '80%'),
            style = {'description_width': 'initial'}
        )
        self.settings.signal_to_noise_ratio = w.BoundedFloatText(
            description = 'SNR: ',
            tooltip = 'Signal to noise ratio',
            value = 3.0,
            min = 0.1,
            max = 100.0,
            step = 0.05,
            disabled = True,
            layout = w.Layout(width = width_percentage_core_widgets),
            style = {'description_width': description_width}
        )
        self.settings.noise_window_size = w.BoundedIntText(
            description = 'NWS: ',
            tooltip = 'Noise window size',
            value = 200,
            min = 10,
            max = 1000,
            step = 1,
            disabled = True,
            layout = w.Layout(width = width_percentage_core_widgets),
            style = {'description_width': description_width}
        )
        self.settings.mean_signal_threshold = w.BoundedFloatText(
            description = 'SAT: ',
            tooltip = 'Signal average threshold',
            value = 10.0,
            min = 0.0,
            max = 255.0,
            step = 0.5,
            disabled = True,
            layout = w.Layout(width = width_percentage_core_widgets),
            style = {'description_width': description_width}
        )
        self.settings.min_peak_count = w.BoundedIntText(
            description = 'MAC: ',
            tooltip = 'Minimum activity counts',
            value = 2,
            min = 0,
            max = 100,
            step = 1,
            disabled = True,
            layout = w.Layout(width = width_percentage_core_widgets),
            style = {'description_width': description_width}
        )
        self.settings.baseline_estimation_method = w.Dropdown(
            description='Baseline estimation method: ',
            value='asls',
            options=[
                ("Asymmetric Least Squares", "asls"),
                ("Fully Automatic Baseline Correction", "fabc"),
                ("Peaked Signal's Asymmetric Least Squares Algorithm", "psalsa"),
                ("Standard Deviation Distribution", "std_distribution")
            ],
            disabled=True,
            layout=w.Layout(width='99%'),
            style={'description_width': 'initial'}
        )
        self.settings.include_variance = w.Checkbox(
            description = 'include variance',
            value = False,
            indent = False,
            disabled = True,
            layout = w.Layout(width = '35%'),
            style = {'description_width': 'initial'}
        )
        self.settings.variance_window_size = w.BoundedIntText(
            description='Variance:',
            disabled=True,
            value=15,
            min=5,
            max=200,
            step=5,
            style={'description_width': 'initial'},
            layout=w.Layout(width='65%', visibility='hidden')
        )
        self.settings.use_frame_range = w.Checkbox(
            description = 'analyze interval',
            indent = False,
            value = False,
            disabled = True,
            layout = w.Layout(width = '35%'),
            style = {'description_width': 'initial'}
        )
        self.settings.frame_interval_to_analyze = w.IntRangeSlider(
            description = 'Frames:',
            disabled = True,
            value = (0, 500),
            min = 0,
            max = 500,
            step = 1,
            style = {'description_width': 'initial'},
            layout = w.Layout(width = '65%', visibility = 'hidden')
        )
        self.settings.customize_octave_filtering = w.Checkbox(
            description = 'configure octaves',
            value = False,
            disabled = True,
            indent = False,
            layout = w.Layout(width = '35%'),
            style = {'description_width': 'initial'}
        )
        self.settings.min_octave_span = w.BoundedFloatText(
            description = 'Min. octaves:',
            tooltip = 'Minimum octaves a ridge needs to span',
            disabled = True,
            value = 1.0,
            min = 0.1,
            max = 30.0,
            step = 0.05,
            style = {'description_width': 'initial'},
            layout = w.Layout(width = '65%', visibility = 'hidden')
        )
        self.settings.save_overview_png = w.Checkbox(
            description = 'Save overview plot',
            value = True,
            disabled = True,
            style = {'description_width': 'initial'}
        )
        self.settings.save_summary_results = w.Checkbox(
            description = 'Save summary results',
            value = True,
            disabled = True,
            style = {'description_width': 'initial'}
        )
        self.settings.save_single_trace_results = w.Checkbox(
            description = 'Save single trace results',
            value = False,
            disabled = True,
            style = {'description_width': 'initial'}
        )
        self.settings.export_to_nwb = w.Checkbox(
            description = 'Export results to NWB',
            value = True,
            disabled = True,
            style = {'description_width': 'initial'}
        )

        # Enable event handling:
        self.settings.include_variance.observe(self._include_variance_config_changed)
        self.settings.use_frame_range.observe(self._limit_analysis_to_interval_changed)
        self.settings.customize_octave_filtering.observe(self._configure_octaves_changed)

        # Arrange elements:
        vbox_core_settings_left = w.VBox(
            [
                self.settings.signal_to_noise_ratio,
                self.vertical_spacer,
                self.settings.mean_signal_threshold
            ],
            layout = w.Layout(width = '50%', align_items = 'flex-start')
        )
        vbox_core_settings_right = w.VBox(
            [
                self.settings.noise_window_size,
                self.vertical_spacer,
                self.settings.min_peak_count
            ],
            layout = w.Layout(width = '50%', align_items = 'flex-end')
        )
        hbox_grid_window_size_settings = w.HBox(
            [
                self.settings.grid_size,
                self.preview_window_size_button
            ],
            layout = w.Layout(width = '100%', justify_content = 'flex-start')
        )
        hbox_core_setting_vboxes = w.HBox(
            [
                vbox_core_settings_left,
                vbox_core_settings_right
            ],
            layout = w.Layout(width = '100%')
        )
        vbox_all_core_settings = w.VBox(
            [
                hbox_grid_window_size_settings,
                self.vertical_spacer,
                hbox_core_setting_vboxes,
                self.vertical_spacer,
                self.settings.baseline_estimation_method
            ],
            layout = w.Layout(width = '90%')
        )
        dashed_separator_line = w.HTML(
            value = "<hr style='border: none; border-bottom: 1px dashed;'>",
            layout = w.Layout(width = '95%')
        )
        optional_info = w.Label(
            value = 'Optional Settings:',
            style = {'text_align': 'left', 'font_weight': 'bold'},
            layout = w.Layout(width = '90%')
        )
        optional_variance_widgets = w.HBox(
            [
                self.settings.include_variance,
                self.settings.variance_window_size
            ],
            layout = w.Layout(width = '100%', align_items = 'flex-start')
        )
        optional_interval_widgets = w.HBox(
            [
                self.settings.use_frame_range,
                self.settings.frame_interval_to_analyze
            ],
            layout = w.Layout(width = '100%', align_items = 'flex-start')
        )
        optional_octave_widgets = w.HBox(
            [
                self.settings.customize_octave_filtering,
                self.settings.min_octave_span
            ],
            layout = w.Layout(width = '100%', align_items = 'flex-start')
        )

        optional_settings = w.VBox(
            [
                optional_variance_widgets,
                self.vertical_spacer,
                optional_interval_widgets,
                self.vertical_spacer,
                optional_octave_widgets
            ],
            layout = w.Layout(width = '90%',
                              align_items = 'flex-start',
                              align_content = 'flex-start',
                              justify_content = 'flex-start')
        )
        results_info = w.Label(
            value = 'Results Settings:',
            style = {'text_align': 'left','font_weight': 'bold'},
            layout = w.Layout(width = '90%')
        )
        results_settings = w.VBox(
            [
                w.HBox(
                    [
                        self.settings.save_overview_png,
                        self.settings.save_summary_results
                    ], layout = w.Layout(width = '100%', align_items = 'flex-start')),
                w.HBox(
                    [
                        self.settings.save_single_trace_results,
                        self.settings.export_to_nwb
                    ], layout = w.Layout(width = '100%', align_items = 'flex-start'))
            ], layout = w.Layout(
                width = '90%',
                align_items = 'flex-start',
                align_content = 'flex-start',
                justify_content = 'center'
            )
        )

        run_analysis_box = w.VBox(
            [self.run_analysis_button],
            layout = w.Layout(width = '100%', align_items = 'center', justify_content = 'center', margin = '10px 0px 0px 0px')
        )
        analysis_settings_box = w.VBox([analysis_settings_info,
                                        vbox_all_core_settings,
                                        self.vertical_spacer,
                                        dashed_separator_line,
                                        optional_info,
                                        optional_settings,
                                        self.vertical_spacer,
                                        dashed_separator_line,
                                        results_info,
                                        results_settings,
                                        self.vertical_spacer,
                                        dashed_separator_line,
                                        run_analysis_box,
                                        self.vertical_spacer],
                                       layout = w.Layout(height = '612px', width = '33%', align_items = 'center', border_top = '1px solid', border_bottom = '1px solid'))
        analysis_settings_box.add_class('box-analysis-settings')
        return analysis_settings_box

    def enable_analysis_settings(self, enable_all_widgets: bool, roi_mode: str) -> None:
        if enable_all_widgets:
            change_widget_state(self.run_analysis_button,
                                disabled=False,
                                button_style='success',
                                tooltip='Click here to start the analysis with all currently specified settings!')
        else:
            change_widget_state(self.run_analysis_button,
                                disabled=True,
                                button_style='')
        for field in self.settings:
            field.disabled = not enable_all_widgets
        self.preview_window_size_button.disabled = not enable_all_widgets

        if roi_mode == 'file':
            self.preview_window_size_button.disabled = True
            self.settings.grid_size.disabled = True


    def _include_variance_config_changed(self, change) -> None:
        if change['name'] == 'value':
            if change['new']:
                self.settings.variance_window_size.layout.visibility = 'visible'
            else:
                self.settings.variance_window_size.layout.visibility = 'hidden'


    def _limit_analysis_to_interval_changed(self, change) -> None:
        if change['name'] == 'value':
            if change['new']:
                self.settings.frame_interval_to_analyze.layout.visibility = 'visible'
            else:
                self.settings.frame_interval_to_analyze.layout.visibility = 'hidden'


    def _configure_octaves_changed(self, change) -> None:
        if change['name'] == 'value':
            if change['new']:
                self.settings.min_octave_span.layout.visibility = 'visible'
            else:
                self.settings.min_octave_span.layout.visibility = 'hidden'

# %% ../nbs/01_view.ipynb 10
class MainScreen:

    def __init__(self) -> None:
        self.logo_path = 'https://raw.githubusercontent.com/jpits30/NeuronActivityTool/master/Logo.png'
        self.output = w.Output()
        self.widget = w.VBox(
            [],
            layout = w.Layout(
                height = '612px',
                width = '67%',
                justify_content = 'center',
                align_items = 'center',
                border_top = '1px solid',
                border_left = '1px solid',
                border_bottom = '1px solid'
            )
        )
        self.widget.add_class('box-main-screen')
        self.current_screen = None

    @property
    def welcome_screen(self) -> w.VBox:
        welcome = w.HTML(
            value="<p style='font-size:32px; font-weight:bold; text-align:center;'>Welcome to</p>",
            layout = w.Layout(width = '99%')
        )
        logo = w.HTML(
            value=f'<img src="{self.logo_path}" width="256" height="256">'
        )
        start_instructions = w.HTML(
            value="<p style='font-size:20px; font-weight:bold; text-align:center;'>Please start by selecting data of a recording to analyze</p>",
            layout = w.Layout(width = '99%')
        )
        return w.VBox([welcome, logo, start_instructions])

    @property
    def output_screen(self) -> w.VBox:
        return w.VBox([self.output])


    def show_welcome_screen(self) -> None:
        self.widget.children = self.welcome_screen.children
        self.current_screen = 'welcome'


    def show_output_screen(self, clear_output: bool = True) -> None:
        if clear_output:
            self.output.clear_output()
        self.widget.children = self.output_screen.children
        self.current_screen = 'output'

# %% ../nbs/01_view.ipynb 11
class WidgetsInterface:

    def __init__(self) -> None:
        self.user_info_panel = UserInfoPanel()
        self.source_data_panel = SourceDataPanel()
        self.analysis_settings_panel = AnalysisSettingsPanel()
        self.main_screen = MainScreen()
        self.nwb_metadata_panel = NWBMetadataPanel()
        self.widget = w.VBox(
            [
                self.source_data_panel.widget,
                w.HBox(
                    [
                        self.analysis_settings_panel.widget,
                        self.main_screen.widget
                    ]
                ),
                self.nwb_metadata_panel.widget,
                self.user_info_panel.widget,
            ],
            layout = w.Layout(align_items = 'stretch', border = '1px solid')
        )
        self.widget.add_class('box-na3-gui')
        self.main_screen.show_welcome_screen()
        self.source_data_panel.settings.batch_mode.observe(self._enable_window_size_widgets)
        self.source_data_panel.settings.focus_area_enabled.observe(self._enable_window_size_widgets)
        self.source_data_panel.settings.roi_mode.observe(self._enable_window_size_widgets)
        self.source_data_panel.settings.data_source_path.register_callback(self._data_source_path_chosen)

    def export_user_settings(self) -> Config:
        """
        Collects all settings from the source data panel and analysis settings panel
        and returns them as a Config object.
        """
        source_data_settings = self.source_data_panel.settings.current_values
        analysis_settings = self.analysis_settings_panel.settings.current_values
        combined_settings = source_data_settings | analysis_settings
        return Config(**combined_settings)

    def export_nwb_metadata(self) -> dict[str, Any]:
        return self.nwb_metadata_panel.metadata.current_metadata

    @property
    def grid_size(self) -> int:
        """
        Returns the grid size value from the analysis settings panel.
        """
        return self.analysis_settings_panel.settings.grid_size.value

    @property
    def data_source_path(self) -> str:
        """
        Returns the data source path from the source data panel.
        """
        return self.source_data_panel.settings.data_source_path.value


    def _data_source_path_chosen(self, file_chooser_obj = None) -> None:
        self.enable_analysis(False)
        if file_chooser_obj.value is not None:
            enable_loading = False
            source_data_path = Path(file_chooser_obj.value)
            if self.source_data_panel.settings.roi_mode.value == 'file':
                if source_data_path.is_dir():
                    enable_loading = True
            else:
                if (self.source_data_panel.settings.batch_mode.value
                        or self.source_data_panel.settings.focus_area_enabled.value):
                    if source_data_path.is_dir():
                        enable_loading = True
                else:
                    if source_data_path.is_file():
                        enable_loading = True
            if enable_loading:
                change_widget_state(self.source_data_panel.load_source_data_button,
                                    disabled = False,
                                    tooltip = 'Click to load the selected source data')
            else:
                change_widget_state(self.source_data_panel.load_source_data_button,
                                    disabled = True,
                                    tooltip = 'Please select which source data to load!')

    def _enable_window_size_widgets(self, change) -> None:
        self.enable_analysis(False)


    def update_infos(self, logs_message: str | None = None, progress_in_percent: float | None = None) -> None:
        if logs_message is not None:
            self.user_info_panel.add_new_logs(logs_message)
        if progress_in_percent is not None:
            self.user_info_panel.update_progress_bar(progress_in_percent)


    def adjust_widgets_to_loaded_data(self, total_frames: int) -> None:
        self.analysis_settings_panel.settings.frame_interval_to_analyze.max = total_frames + 1
        self.analysis_settings_panel.settings.frame_interval_to_analyze.value = (1, total_frames + 1)


    def enable_analysis(self, enable: bool = True) -> None:
        roi_mode = self.source_data_panel.settings.roi_mode.value
        self.analysis_settings_panel.enable_analysis_settings(enable, roi_mode)



"""Defines the import for different source filetypes."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_input.ipynb.

# %% auto 0
__all__ = ['FocusAreaPathRestrictions', 'Data', 'Recording', 'ROI', 'DataLoader', 'GridWrapperROILoader', 'RecordingLoader',
           'AVILoader', 'TIFFLoader', 'NWBRecordingLoader', 'ROILoader', 'ImageJROILoader', 'NWBROILoader',
           'DataLoaderFactory', 'RecordingLoaderFactory', 'ROILoaderFactory',
           'get_filepaths_with_supported_extension_in_dirpath', 'RecLoaderROILoaderCombinator',
           'test_unsupported_file_extension', 'test_successful_recording_loading', 'test_successful_roi_loading']

# %% ../nbs/03_input.ipynb 3
import imageio.v3 as iio
from pathlib import Path
import numpy as np
from shapely import Polygon
import roifile
from skimage.draw import polygon
from skimage.measure import find_contours
from pynwb import NWBHDF5IO

from abc import ABC, abstractmethod
from typing import List, Dict, Tuple, Union, Optional, Any

# %% ../nbs/03_input.ipynb 4
class FocusAreaPathRestrictions:
    
    @property
    def supported_dir_names(self) -> List[str]:
        supported_dir_names =  ['focus_area', 'focus_areas', 'focus-area', 'focus-areas', 'focus area', 'focus areas',
                                'Focus_Area', 'Focus_Areas', 'Focus-Area', 'Focus-Areas', 'Focus Area', 'Focus Areas']
        return supported_dir_names

# %% ../nbs/03_input.ipynb 6
class Data(ABC):

    @abstractmethod
    def _parse_loaded_data(self, loaded_data: Any) -> None:
        pass

    def __init__(self, filepath: Path, loaded_data: Any) -> None:
        self.filepath = filepath
        self._parse_loaded_data(loaded_data)

# %% ../nbs/03_input.ipynb 7
class Recording(Data):

    def _parse_loaded_data(self, loaded_data: np.ndarray) -> None:
        self.zstack = loaded_data
        self.estimated_bit_depth = self._estimate_bit_depth()
        self.preview = self._create_brightness_and_contrast_enhanced_preview()


    def _estimate_bit_depth(self) -> int:
        max_bit_value = self.zstack.max()
        if max_bit_value <= 255:
            estimated_bit_depth = 255
        elif max_bit_value <= 4095:
            estimated_bit_depth = 4095
        elif max_bit_value <= 65535:
            estimated_bit_depth = 65535
        else:
            raise ValueError(f'Max bit value in recording found to be {max_bit_value}, but NA3 currently only handles up to 16-bit recordings!')
        return estimated_bit_depth


    def _create_brightness_and_contrast_enhanced_preview(self, percentile_for_adjustment: int=1) -> np.ndarray:
        raw_image = self.zstack[0, :, :, :].copy() # ensure that dimensions are the same as for ".get_single_frame_as_preview()"
        lower_percentile_bit_value = np.percentile(raw_image, percentile_for_adjustment)
        upper_percentile_bit_value = np.percentile(raw_image, 100-percentile_for_adjustment)
        contrast_adjustment_factor = self.estimated_bit_depth / (upper_percentile_bit_value - lower_percentile_bit_value)
        brightness_adjustment_factor = -(contrast_adjustment_factor * lower_percentile_bit_value)
        raw_image_clipped_at_percentile_bit_values = self._clip_image_at_bit_values(raw_image, lower_percentile_bit_value, upper_percentile_bit_value)
        brightness_contrast_adjusted_image = contrast_adjustment_factor * raw_image_clipped_at_percentile_bit_values + brightness_adjustment_factor
        return brightness_contrast_adjusted_image
        

    def _compute_contrast_and_brightness_adjustment_factors(self, raw_image: np.ndarray, percentile_for_adjustment: int=1) -> Tuple[float, float]:
        lower_percentile_bit_value = np.percentile(raw_image, percentile_for_adjustment)
        upper_percentile_bit_value = np.percentile(raw_image, 100-percentile_for_adjustment)
        contrast_adjustment = self.estimated_bit_depth / (upper_percentile_bit_value - lower_percentile_bit_value)
        brightness_adjustment = -(contrast_adjustment * lower_percentile_bit_value)
        return contrast_adjustment, brightness_adjustment


    def _clip_image_at_bit_values(self, raw_image: np.ndarray, min_bit_value: float, max_bit_value: float) -> np.ndarray:
        raw_image[raw_image <= min_bit_value] = min_bit_value
        raw_image[raw_image >= max_bit_value] = max_bit_value
        return raw_image

# %% ../nbs/03_input.ipynb 8
class ROI(Data):

    def _parse_loaded_data(self, loaded_data: List[Tuple[int, int]]) -> None:
        self.boundary_row_col_coords = loaded_data
        self.as_polygon = self._convert_to_valid_polygon()


    def _convert_to_valid_polygon(self) -> Polygon:
        roi_as_polygon = Polygon(self.boundary_row_col_coords)
        # There was an inconsistency discovered when trying to convert ROIs exported from ImageJ into valid Polygons in shapely.
        # Sometimes, the same ROIs will export in a way that allows to create valid Polygons, and sometimes the same ROIs will be 
        # exported in a way that they can´t be converted into valid Polygons. As of now (16.12.2025), the reason for this inconsistency
        # remained unclear. Nonetheless, valid Polygons or not - they always lead to the identical results when used via na3 for processing.
        # Thus, we decided to only log the result of this check, but no longer raise an error that aborts the processing.
        if not roi_as_polygon.is_valid:
            message = (f'Shapely could not create a valid Polygon out of your ROI: {self.filepath}. In our experience, this does not impact the results. '
                       'However, we recommend you try to re-create the ROIs and re-run the processing with NA3. We noticed conversion inconsistencies '
                       'particularly for ROIs exported from ImageJ2 / FIJI. Exporting them again typically resolves this warning.') 
            print(message)
        return roi_as_polygon


    def add_label_id(self, label_id: str) -> None:
        assert type(label_id) == str, f'"label_id" must be a string. However, you passed {label_id} which is of type {type(label_id)}.'
        setattr(self, 'label_id', label_id)


    def create_nwb_compliant_pixel_mask(self):
        boundary_coords_with_subpixel_precision = np.asarray(self.boundary_row_col_coords)
        row_idxs, col_idxs = polygon(boundary_coords_with_subpixel_precision[:, 0], boundary_coords_with_subpixel_precision[:, 1])
        weights = np.ones(row_idxs.shape[0], dtype='int64')
        self.pixel_mask = np.stack((row_idxs, col_idxs, weights), axis = 1)

# %% ../nbs/03_input.ipynb 10
class DataLoader(ABC):

    @abstractmethod
    def load_and_parse_file_content(self) -> Union[Data, List[Data]]:
        # This method will be called when the data should be loaded for analysis
        pass

    def __init__(self, filepath: Path) -> None:
        self.filepath = filepath

# %% ../nbs/03_input.ipynb 11
class GridWrapperROILoader(DataLoader):

    def set_configs_for_grid_creation(self, image_width: int, image_height: int, window_size: int) -> None:
        self.configs = {}
        self._add_to_configs_and_create_as_attribute('image_width', image_width)
        self._add_to_configs_and_create_as_attribute('image_height', image_height)
        self._add_to_configs_and_create_as_attribute('window_size', window_size)


    def _add_to_configs_and_create_as_attribute(self, attribute_name: str, value: Any) -> None:
        self.configs[attribute_name] = value
        setattr(self, attribute_name, self.configs[attribute_name])

    
    def load_and_parse_file_content(self) -> List[ROI]:
        row_cropping_idx, col_cropping_idx = self._get_cropping_indices_to_adjust_for_window_size()
        self._add_to_configs_and_create_as_attribute('row_cropping_idx', row_cropping_idx)
        self._add_to_configs_and_create_as_attribute('col_cropping_idx', col_cropping_idx)        
        grid_row_idxs, grid_col_idxs = self._get_row_col_idxs_of_grid()
        grid_row_labels, grid_col_labels = self._get_row_col_labels_for_rois_in_grid()
        self._add_to_configs_and_create_as_attribute('max_len_row_label_id', len(str(grid_row_labels[-1])))
        self._add_to_configs_and_create_as_attribute('max_len_col_label_id', len(str(grid_col_labels[-1])))
        all_rois = []
        for row_idx, row_label in zip(grid_row_idxs, grid_row_labels):
            for col_idx, col_label in zip(grid_col_idxs, grid_col_labels):
                square_corner_row_col_coords = self._get_boundary_row_col_coords_single_square(row_idx, col_idx)
                label_id = f'{row_label}/{col_label}'
                square_roi = ROI(self.filepath, square_corner_row_col_coords)
                square_roi.add_label_id(label_id)
                all_rois.append(square_roi)
        return all_rois     

    
    def _get_cropping_indices_to_adjust_for_window_size(self) -> Tuple[int, int]:
        row_cropping_index = (self.image_height // self.window_size) * self.window_size
        col_cropping_index = (self.image_width // self.window_size) * self.window_size
        return row_cropping_index, col_cropping_index
        

    def _get_row_col_idxs_of_grid(self) -> Tuple[np.ndarray, np.ndarray]:
        grid_row_idxs = np.arange(start = 0, stop = self.row_cropping_idx, step = self.window_size)
        grid_col_idxs = np.arange(start = 0, stop = self.col_cropping_idx, step = self.window_size)
        return grid_row_idxs, grid_col_idxs


    def _get_row_col_labels_for_rois_in_grid(self) -> Tuple[np.ndarray, np.ndarray]:
        grid_row_labels = np.arange(start = 1, stop = self.row_cropping_idx / self.window_size + 1, step = 1, dtype = 'int')
        grid_col_labels = np.arange(start = 1, stop = self.col_cropping_idx / self.window_size + 1, step = 1, dtype = 'int')
        return grid_row_labels, grid_col_labels
                
    
    def _get_boundary_row_col_coords_single_square(self, upper_left_corner_row_idx: int, upper_left_corner_col_idx: int) -> List[Tuple[(int, int)]]:
        upper_left_corner = (upper_left_corner_row_idx, upper_left_corner_col_idx)
        upper_right_corner = (upper_left_corner_row_idx, upper_left_corner_col_idx + self.window_size)
        lower_right_corner = (upper_left_corner_row_idx + self.window_size, upper_left_corner_col_idx + self.window_size)
        lower_left_corner = (upper_left_corner_row_idx + self.window_size, upper_left_corner_col_idx)
        return [upper_left_corner, lower_left_corner, lower_right_corner, upper_right_corner, upper_left_corner]

# %% ../nbs/03_input.ipynb 12
class RecordingLoader(DataLoader):


    @abstractmethod
    def _get_all_frames(self) -> np.ndarray: 
        # To be implemented in individual subclasses
        # Shape of returned numpy array: [frames, rows, cols] or [frames, rows, cols, color_channels]
        pass


    def _load_all_frames(self) -> np.ndarray: 
        all_frames = self._get_all_frames()
        all_frames = self._ensure_array_is_4d(all_frames)
        if self._check_if_multiple_color_channels_exist(all_frames):
            if self._check_if_color_channels_are_redundant(all_frames):
                all_frames = self._convert_to_grayscale(all_frames)
            else:
                raise ValueError('The color channels of the recording you attempted to load are incorrect. Currently, only single '
                                 f'channel or redundant RGB (i.e. 1 or 3 color channels) are supported. However, your data has: {all_frames.shape[3]}.')
        return all_frames


    def load_and_parse_file_content(self) -> Recording:
        all_frames = self._load_all_frames()
        recording = Recording(self.filepath, all_frames)
        return recording


    def _ensure_array_is_4d(self, zstack: np.ndarray) -> np.ndarray:
        if len(zstack.shape) < 3:
            raise ValueError('The shape of the zstack numpy array is not correct. It should be at least 3 dimensional [frames, rows, cols], '
                             f'with an optional fourth dimension of color channels. However, the current shape is: {zstack.shape}.')
        elif len(zstack.shape) == 3:
            zstack = np.expand_dims(zstack, axis=3)
        elif len(zstack.shape) > 4:
            raise ValueError('The shape of the zstack numpy array is not correct. It can´t be more than 4 dimensional: '
                             f'[frames, rows, cols, colors]. However, the current shape is: {zstack.shape}.')
        return zstack
        

    def _check_if_multiple_color_channels_exist(self, zstack: np.ndarray) -> bool:
        return zstack.shape[3] > 1
        
    
    def _check_if_color_channels_are_redundant(self, zstack: np.ndarray) -> bool:
        reference_channel_idx = 0
        color_channels_are_equal = []
        for idx_of_channel_to_compare in range(1, zstack.shape[3]):
            if np.array_equal(zstack[:, :, :, reference_channel_idx], zstack[: , :, :, idx_of_channel_to_compare]) == True:
                color_channels_are_equal.append(True)
            else:
                color_channels_are_equal.append(False)
                break
        return all(color_channels_are_equal)


    def _convert_to_grayscale(self, zstack: np.ndarray) -> np.ndarray:
        return zstack[:, :, :, 0:1]

# %% ../nbs/03_input.ipynb 13
class AVILoader(RecordingLoader):

    def _get_all_frames(self) -> np.ndarray: 
        return iio.imread(self.filepath)

# %% ../nbs/03_input.ipynb 14
class TIFFLoader(RecordingLoader):

    def _get_all_frames(self) -> np.ndarray: 
        return iio.imread(self.filepath)

# %% ../nbs/03_input.ipynb 15
class NWBRecordingLoader(RecordingLoader):

    def _get_all_frames(self) -> np.ndarray:
        with NWBHDF5IO(self.filepath, "r+") as io:
            nwbfile = io.read()
            if 'OnePhotonSeries' in nwbfile.acquisition.keys():
                all_frames = nwbfile.acquisition['OnePhotonSeries'].data[:]
            elif 'TwoPhotonSeries' in nwbfile.acquisition.keys():
                all_frames = nwbfile.acquisition['TwoPhotonSeries'].data[:]
            else:
                raise ValueError('The NWB file you try to load does not have a recording stored as "OnePhotonSeries" or "TwoPhotonSeries" under Acquisition.')
        if len(all_frames.shape) < 4:
            all_frames = all_frames[..., np.newaxis]        
        return all_frames

# %% ../nbs/03_input.ipynb 16
class ROILoader(DataLoader):

    @abstractmethod
    def _get_boundary_row_col_coords_for_all_rois_in_source_data(self) -> List[List[Tuple[int, int]]]: 
        # To be implemented in individual subclasses
        # Return a list of Tuples, where each tuple represents one boundary point: (row_coord, col_coord)
        pass

    
    def load_and_parse_file_content(self) -> List[ROI]:
        boundary_row_col_coords_for_all_rois = self._get_boundary_row_col_coords_for_all_rois_in_source_data()
        all_rois = []
        for boundary_row_col_coords_single_roi in boundary_row_col_coords_for_all_rois:
            boundary_row_col_coords_single_roi = self._add_first_boundary_point_also_add_end_to_close_roi(boundary_row_col_coords_single_roi)
            roi = ROI(self.filepath, boundary_row_col_coords_single_roi)
            all_rois.append(roi)
        return all_rois


    def _add_first_boundary_point_also_add_end_to_close_roi(self, boundary_row_col_coords: List[Tuple[int, int]]) -> List[Tuple[int, int]]:
        first_boundary_point_coords = boundary_row_col_coords[0]
        boundary_row_col_coords.append(first_boundary_point_coords)
        return boundary_row_col_coords

# %% ../nbs/03_input.ipynb 17
class ImageJROILoader(ROILoader):

    
    def _get_boundary_row_col_coords_for_all_rois_in_source_data(self) -> List[List[Tuple[int, int]]]:
        roi_file_content = roifile.roiread(self.filepath)
        if type(roi_file_content) == list:
            all_rois = self._extract_boundary_row_col_coords_from_roi_set(roi_file_content)
        else:
            all_rois = [self._extract_boundary_row_col_coords_from_single_roi(roi_file_content)]
        return all_rois


    def _extract_boundary_row_col_coords_from_roi_set(self, all_imagej_rois: List[roifile.roifile.ImagejRoi]) -> List[List[Tuple[int, int]]]:
        boundary_coords_all_rois = []
        for imagej_roi in all_imagej_rois:
            boundary_row_col_coords_single_roi = self._extract_boundary_row_col_coords_from_single_roi(imagej_roi)
            boundary_coords_all_rois.append(boundary_row_col_coords_single_roi)
        return boundary_coords_all_rois
    
        
    def _extract_boundary_row_col_coords_from_single_roi(self, imagej_roi: roifile.roifile.ImagejRoi) -> List[Tuple[int, int]]:
        row_coords = imagej_roi.coordinates()[:, 1]
        col_coords = imagej_roi.coordinates()[:, 0]
        boundary_row_col_coords = list(zip(row_coords, col_coords))
        return boundary_row_col_coords

# %% ../nbs/03_input.ipynb 18
class NWBROILoader(ROILoader):

    def _get_boundary_row_col_coords_for_all_rois_in_source_data(self) -> List[List[Tuple[int, int]]]:
        with NWBHDF5IO(self.filepath, "r") as io:
            nwbfile = io.read()
            recording_frame_shape = nwbfile.processing['ophys']['ImageSegmentation']['PlaneSegmentation'].reference_images[0].data.shape[1:3]
            rois_df = nwbfile.processing['ophys']['ImageSegmentation']['PlaneSegmentation'].to_dataframe()
        all_rois = []
        for roi_id in rois_df.index.values:
            pixel_mask = rois_df.at[roi_id, 'pixel_mask']
            binary_mask = np.zeros(recording_frame_shape, dtype='uint8')
            for x, y, _ in pixel_mask:
                binary_mask[int(x), int(y)] = 1
            contour_results = find_contours(binary_mask)
            if len(contour_results) != 1:
                raise ValueError(f'Expected to find exactly one contour, but found {len(contour_results)} for ROI ID')
            boundary_row_col_coords = [(float(row), float(col)) for row, col in contour_results[0]]
            all_rois.append(boundary_row_col_coords)
        return all_rois

# %% ../nbs/03_input.ipynb 20
class DataLoaderFactory(ABC):

    @property
    @abstractmethod
    def supported_extensions_per_data_loader(self) -> Dict[DataLoader, List[str]]:
        pass

    
    @property
    def all_supported_extensions(self) -> List[str]:
        all_supported_extensions = []
        for value in self.supported_extensions_per_data_loader.values():
            all_supported_extensions += value
        return all_supported_extensions

    
    def get_loader(self, filepath: Path) -> DataLoader:
        self._assert_validity_of_filepath(filepath)
        data_loader = self._get_loader_for_file_extension(filepath)
        return data_loader
        

    def _assert_validity_of_filepath(self, filepath: Path) -> None:
        assert isinstance(filepath, Path), f'filepath must be an instance of a pathlib.Path. However, you passed {filepath}, which is of type {type(filepath)}'
        assert filepath.exists(), f'The filepath you provided ({filepath}) does not seem to exist!'
        

    def _get_loader_for_file_extension(self, filepath: Path) -> DataLoader:
        matching_loader = None
        for loader_subclass, supported_extensions in self.supported_extensions_per_data_loader.items():
            if filepath.suffix in supported_extensions:
                matching_loader = loader_subclass(filepath)
                break
        if matching_loader == None:
            raise NotImplementedError('It seems like there is no DataLoader implemented for the specific filetype you´re trying to load - sorry!')
        return matching_loader

# %% ../nbs/03_input.ipynb 21
class RecordingLoaderFactory(DataLoaderFactory):

    @property
    def supported_extensions_per_data_loader(self) -> Dict[RecordingLoader, List[str]]:
        supported_extensions_per_data_loader = {
            AVILoader: ['.avi'],
            TIFFLoader: ['.tif', '.tiff'],
            NWBRecordingLoader: ['.nwb']
        }
        return supported_extensions_per_data_loader

# %% ../nbs/03_input.ipynb 22
class ROILoaderFactory(DataLoaderFactory):

    @property
    def supported_extensions_per_data_loader(self) -> Dict[ROILoader, List[str]]:
        supported_extensions_per_data_loader = {
            ImageJROILoader: ['.roi', '.zip'],
            NWBROILoader: ['.nwb']
        }
        return supported_extensions_per_data_loader

# %% ../nbs/03_input.ipynb 23
def get_filepaths_with_supported_extension_in_dirpath(dirpath: Path, all_supported_extensions: List[str], max_results: Optional[int]=None) -> List[Path]:
    all_filepaths_with_supported_extension = []
    for elem in dirpath.iterdir():
        if elem.is_file() and elem.suffix in all_supported_extensions and not elem.name.startswith(('.', '._')):
            all_filepaths_with_supported_extension.append(elem)
    if type(max_results) == int:
        assert len(all_filepaths_with_supported_extension) <= max_results, (
            f'There are more than {max_results} file(s) of supported type in {dirpath}, '
            f'but only a maximum of {max_results} are allowed. Please remove at least '
            f'{len(all_filepaths_with_supported_extension) - max_results} of the following' 
            f'files and try again: {all_filepaths_with_supported_extension}'
        )
    return all_filepaths_with_supported_extension

# %% ../nbs/03_input.ipynb 24
class RecLoaderROILoaderCombinator:

        
    def __init__(self, dir_path: Path) -> None:
        self.dir_path = dir_path

    
    def get_all_recording_and_roi_loader_combos(self) -> List[Tuple[RecordingLoader, ROILoader]]:
        recording_loader = self._get_the_recording_loader()
        all_roi_loaders = self._get_all_roi_loaders()
        if len(all_roi_loaders) > 0:
            rec_roi_loader_combos = [(recording_loader, roi_loader) for roi_loader in all_roi_loaders]
        else:
            rec_roi_loader_combos = [(recording_loader, None)]
        return rec_roi_loader_combos
    

    def _get_the_recording_loader(self) -> RecordingLoader:
        recording_loader_factory = RecordingLoaderFactory()
        recording_filepath = get_filepaths_with_supported_extension_in_dirpath(self.dir_path, recording_loader_factory.all_supported_extensions, 1)[0]
        recording_loader = recording_loader_factory.get_loader(recording_filepath)
        return recording_loader


    def _get_all_roi_loaders(self) -> List[ROILoader]:
        roi_loader_factory = ROILoaderFactory()
        all_roi_filepaths = get_filepaths_with_supported_extension_in_dirpath(self.dir_path, roi_loader_factory.all_supported_extensions)
        all_roi_loaders = [roi_loader_factory.get_loader(filepath) for filepath in all_roi_filepaths]
        return all_roi_loaders

# %% ../nbs/03_input.ipynb 26
def test_unsupported_file_extension(loader_factory: DataLoaderFactory, filepath: Path) -> bool:
    try:
        loader_factory.get_loader(filepath)
    except NotImplementedError:
        return True
    else:
        return False

# %% ../nbs/03_input.ipynb 27
def test_successful_recording_loading(filepath: Path) -> bool:
    recording_loader_factory = RecordingLoaderFactory()
    recording_loader = recording_loader_factory.get_loader(filepath)
    return isinstance(recording_loader.load_and_parse_file_content(), Recording)

# %% ../nbs/03_input.ipynb 28
def test_successful_roi_loading(filepath: Path) -> bool:
    roi_loader_factory = ROILoaderFactory()
    roi_loader = roi_loader_factory.get_loader(filepath)
    parsed_rois = roi_loader.load_and_parse_file_content()
    if type(parsed_rois) == list:
        are_rois = [isinstance(roi, ROI) for roi in parsed_rois]
        if all(are_rois):
            return True
        else:
            return False
    else:
        return False



"""Defines the structure of individual analysis jobs."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_processing.ipynb.

# %% auto 0
__all__ = ['process_analysis_rois', 'AnalysisJob']

# %% ../nbs/04_processing.ipynb 3
import multiprocessing
import numpy as np
import pandas as pd
from pathlib import Path
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes._axes import Axes 
from datetime import datetime
from shapely import get_coordinates


from dateutil.tz import tzlocal
from uuid import uuid4
from pynwb import NWBHDF5IO, NWBFile
from pynwb.ophys import (
    Fluorescence,
    ImageSegmentation,
    OnePhotonSeries,
    OpticalChannel,
    RoiResponseSeries
)
from typing import Any

# %% ../nbs/04_processing.ipynb 4
from .input import DataLoader, ROI, GridWrapperROILoader
from .datamodels import Config
from .analysis import AnalysisROI
from . import results, __version__

# %% ../nbs/04_processing.ipynb 5
def process_analysis_rois(analysis_roi: AnalysisROI, config: Config) -> AnalysisROI:
    analysis_roi.compute_mean_intensity_timeseries(config.use_frame_range, config.start_frame_idx, config.end_frame_idx)
    if np.mean(analysis_roi.mean_intensity_over_time) >= config.mean_signal_threshold:
        analysis_roi.detect_peaks(config.signal_to_noise_ratio, config.min_octave_span, config.noise_window_size)
        analysis_roi.estimate_baseline(config.baseline_estimation_method)
        analysis_roi.compute_area_under_curve()
        analysis_roi.compute_amplitude_and_delta_f_over_f()
        analysis_roi.compute_variance_area(config.variance_window_size)
    return analysis_roi

# %% ../nbs/04_processing.ipynb 6
class AnalysisJob:

    def __init__(self, 
                 number_of_parallel_processes: int,
                 data_loaders: dict[str, DataLoader | list[DataLoader]],
                 results_dir: Path | None = None
                ) -> None:
        self.number_of_parallel_processes = number_of_parallel_processes
        self.recording_loader = data_loaders['recording']
        self.results_dir = results_dir
        self.results_dir_path: Path | None = None
        self.nwb_metadata: dict | None = None
        self.analysis_start_datetime: datetime | None = None
        if 'rois' in data_loaders.keys():
            self.rois_source = 'file'
            self.roi_loaders = data_loaders['rois']
        else:
            self.rois_source = 'grid'
            self.roi_loaders = [GridWrapperROILoader(self.recording_loader.filepath)]
        if 'focus_area' in data_loaders.keys():
            self.focus_area_enabled = True
            self.focus_area_loader = data_loaders['focus_area']
        else:
            self.focus_area_enabled = False
            self.focus_area = None

    
    def preview_window_size(self, window_size: int) -> tuple[Figure, Axes]:
        self.load_data_into_memory(window_size)
        grid_configs = self.roi_loaders[0].configs
        fig, ax = results.plot_window_size_preview(self.recording.preview, grid_configs, self.focus_area)
        return fig, ax

    
    def load_data_into_memory(self, window_size: int) -> None:
        self.recording = self.recording_loader.load_and_parse_file_content()
        if not hasattr(self, 'all_analysis_rois'):
            self.all_analysis_rois = self._create_all_analysis_rois(window_size)
        elif self.rois_source == 'grid':
            if self.roi_loaders[0].window_size != window_size:
                self.all_analysis_rois = self._create_all_analysis_rois(window_size)


    def _create_all_analysis_rois(self, window_size: int) -> list[AnalysisROI]:     
        self.all_rois = self._load_data_from_all_roi_loaders(window_size)
        if self.focus_area_enabled:
            self.focus_area = self.focus_area_loader.load_and_parse_file_content()[0]
            self.all_rois = self._filter_rois_by_focus_area()
        return self._create_analysis_rois()

    
    def _load_data_from_all_roi_loaders(self, window_size: int) -> list[ROI]:
        all_rois = []
        for roi_loader in self.roi_loaders:
            if type(roi_loader) == GridWrapperROILoader:
                roi_loader.set_configs_for_grid_creation(self.recording.preview.shape[1], self.recording.preview.shape[0], window_size)
            all_rois += roi_loader.load_and_parse_file_content()
        if self.rois_source == 'grid':
            self.grid_configs = self.roi_loaders[0].configs
        else:
            self.grid_configs = None
        return all_rois
    

    def _filter_rois_by_focus_area(self) -> list[ROI]:
        filtered_rois = [roi for roi in self.all_rois if roi.as_polygon.within(self.focus_area.as_polygon)]
        return filtered_rois
        

    def _create_analysis_rois(self) -> list[AnalysisROI]:
        all_analysis_rois = []
        if self.rois_source == 'file':
            self._create_and_add_label_ids_to_all_rois_from_file()
        for roi in self.all_rois:
            roi_bounding_box_row_col_coords = get_coordinates(roi.as_polygon.envelope).astype('int')
            row_min, col_min = roi_bounding_box_row_col_coords.min(axis=0)
            row_max, col_max = roi_bounding_box_row_col_coords.max(axis=0)   
            zstack = self.recording.zstack[:, row_min:row_max, col_min:col_max, :]
            analysis_roi = AnalysisROI(roi, (row_min, col_min), zstack)
            all_analysis_rois.append(analysis_roi)
        return all_analysis_rois


    def _create_and_add_label_ids_to_all_rois_from_file(self) -> None:
        roi_count = len(self.all_rois)
        zfill_factor = int(np.log10(roi_count)) + 1
        for idx, roi in enumerate(self.all_rois):
            label_id = str(idx + 1).zfill(zfill_factor)
            roi.add_label_id(label_id)

    def _create_results_dir(self) -> Path:
        """
        Creates and returns the directory path where analysis results will be saved.

        The directory name is constructed using the analysis start datetime, the recording filename, and optionally the focus area filename if focus area is enabled. If a custom results directory was provided during initialization, it is used instead. The method ensures the directory exists (creates it if necessary) and returns the Path object for the results directory.
        """
        prefix_with_datetime = self.analysis_start_datetime.strftime('%Y_%m_%d_%H-%M-%S_results_for')
        recording_filename = self.recording.filepath.stem
        if self.focus_area_enabled:
            focus_area_filename = self.focus_area.filepath.stem
            results_dir_name = f'{prefix_with_datetime}_{recording_filename}_with_{focus_area_filename}'
        else:
            results_dir_name = f'{prefix_with_datetime}_{recording_filename}'
        if self.results_dir is not None:
            results_dir_path = self.results_dir / results_dir_name
        else:
            results_dir_path = self.recording.filepath.parent / results_dir_name
        results_dir_path.mkdir(exist_ok=True, parents=True)
        return results_dir_path

    
    def run_analysis(self, config: Config) -> None:
        self.analysis_start_datetime = datetime.now(tzlocal())
        self.load_data_into_memory(config.grid_size)
        self.results_dir_path = self._create_results_dir()
        copy_of_all_analysis_rois = self.all_analysis_rois.copy()
        with multiprocessing.Pool(processes = self.number_of_parallel_processes) as pool:
            processed_analysis_rois = pool.starmap(process_analysis_rois, [(analysis_roi, config) for analysis_roi in copy_of_all_analysis_rois])
        self.all_analysis_rois = processed_analysis_rois


    def create_results(self, config: Config, nwb_metadata: dict[str, Any] = None) -> None:
        activity_filtered_analysis_rois = [roi for roi in self.all_analysis_rois if roi.peaks_count >= config.min_peak_count]
        self.activity_overview_plot = results.plot_activity_overview(analysis_rois_with_sufficient_activity = activity_filtered_analysis_rois,
                                                                     preview_image = self.recording.preview,
                                                                     indicate_activity = True,
                                                                     focus_area = self.focus_area,
                                                                     grid_configs = self.grid_configs)
        if config.save_overview_png:
            self.activity_overview_plot[0].savefig(self.results_dir_path.joinpath('activity_overview.png'), dpi=300)
            label_id_overview_fig, label_id_overview_ax = results.plot_rois_with_label_id_overview(
                analysis_rois_with_sufficient_activity=activity_filtered_analysis_rois,
                preview_image=self.recording.preview,
                focus_area=self.focus_area,
                grid_configs=self.grid_configs)
            label_id_overview_fig.savefig(self.results_dir_path.joinpath('ROI_label_IDs_overview.png'), dpi=300)
            plt.close()
        if config.save_summary_results:
            self._create_and_save_csv_result_files(activity_filtered_analysis_rois)
            self._create_and_save_individual_traces_pdf_result_file(activity_filtered_analysis_rois)
        if config.save_single_trace_results:
            self._create_and_save_single_trace_results_as_csv(activity_filtered_analysis_rois)
        if config.export_to_nwb:
            self._export_all_data_to_nwb(nwb_metadata)
 
    def _export_all_data_to_nwb(self, nwb_metadata: dict[str, Any] = None) -> None:
        if nwb_metadata is None:
            nwb_metadata = {}
        nwb_file = self._initialize_nwb_file(nwb_metadata)
        nwb_file = self._add_source_data(nwb_file, nwb_metadata)
        nwb_file = self._add_na3_results(nwb_file)
        self._write_nwb_file(nwb_file)


    def _initialize_nwb_file(self, nwb_metadata: dict[str, Any]) -> NWBFile:
        nwb_file = NWBFile(
            session_description=f'Autogenerated NWB file by neuralactivitycubic version {__version__}',
            identifier=str(uuid4()),
            session_start_time=datetime.now(tzlocal()),
            experimenter=nwb_metadata.get('experimenter', ''),
            experiment_description=nwb_metadata.get('experiment_description', ''),
            lab=nwb_metadata.get('lab', ''),
            institution=nwb_metadata.get('institution', ''),
            notes=nwb_metadata.get('notes', ''),
        )
        return nwb_file


    def _add_source_data(self, nwb_file: NWBFile, nwb_metadata: dict[str, Any]) -> NWBFile:
        # Add recording source data
        microscope = nwb_file.create_device(
            name = 'Microscope',
            model_number=nwb_metadata.get('model_number', ''),
            manufacturer=nwb_metadata.get('manufacturer', ''),
            description=nwb_metadata.get('description', ''),
        )
        optical_channel = OpticalChannel(
            name = 'channel_01',
            description = nwb_metadata.get('imaging_description', ''),
            emission_lambda = nwb_metadata.get('emission_wavelength', 0.0),
        )
        imaging_plane = nwb_file.create_imaging_plane(
            name = 'ImagingPlane',
            optical_channel = optical_channel,
            description = 'In vitro Calcium Imaging of neuronal cell culture',
            device = microscope,
            excitation_lambda = nwb_metadata.get('excitation_wavelength', 0.0),
            indicator = nwb_metadata.get('calcium_indicator', ''),
            location = nwb_metadata.get('location', ''),
        )
        cai_series = OnePhotonSeries(
            name = 'OnePhotonSeries',
            imaging_plane = imaging_plane,
            unit = 'a.u.',
            data = self.recording.zstack[:, :, :, 0], 
            rate = 10.0 #tbd
        )
        nwb_file.add_acquisition(cai_series)
        # Add ROI source data:
        ophys_module = nwb_file.create_processing_module(
            name="ophys", description="optical physiology processed data"
        )
        img_seg = ImageSegmentation()
        roi_plane_seg = img_seg.create_plane_segmentation(
            name="PlaneSegmentation",
            description="ROIs for analysis; created manually using ImageJ2",
            imaging_plane=imaging_plane,
            reference_images=cai_series,
        )
        ophys_module.add(img_seg)
        roi_plane_seg.add_column(name = 'ROI_label_ID', description = 'ROI label ID assigned by neuralactivitycubic', index = False)
        for roi in self.all_rois:
            roi.create_nwb_compliant_pixel_mask()
            roi_plane_seg.add_roi(pixel_mask = roi.pixel_mask, ROI_label_ID = roi.label_id)
        return nwb_file


    def _add_na3_results(self, nwb_file: NWBFile) -> NWBFile:
        # Add mean intensity time series per ROI:
        plane_segmentation = nwb_file.processing['ophys']['ImageSegmentation']['PlaneSegmentation']
        all_rois_table_region = plane_segmentation.create_roi_table_region(description = 'All ROIs')
        stacked_arrays = np.stack([analysis_roi.mean_intensity_over_time for analysis_roi in self.all_analysis_rois], axis=1)
        mean_int_roi_response_series = RoiResponseSeries(
            name="RoiResponseSeries",
            description="Mean intensity time series for all ROIs",
            data=stacked_arrays,
            rois=all_rois_table_region,
            unit="bit-values",
            rate=10.0, # use rate of image series in acquisition of NWB file
        )
        mean_int_fluorescence_series = Fluorescence(mean_int_roi_response_series)
        nwb_file.processing['ophys'].add(mean_int_fluorescence_series)
        # Add all analysis results
        custom_unit_column_infos = [
            ('ROI_label_ID', 'ROI label ID assigned by neuralactivitycubic', False),
            ('total_peaks_count', 'Total number of detected peaks', False),
            ('variance_area', 'Computed variance area', False),
            ('peaks_frame_idxs', 'The frame indices of all peaks', True),
            ('peaks_bit_value', 'The fluorescence intensities (bit values) of all peak', True),
            ('peaks_amplitude', 'The amplitudes of all peaks (above the computed baseline)', True),
            ('peaks_df_over_f', 'The delta-F over F values of all peaks', True),
            ('peaks_auc', 'The area under curve values of all peaks (np.NaN for individual peaks if no AUC could be computed)', True),
            ('peaks_classification', 'The classification of all peaks as assigned by neuralactivitycubic', True)
        ]
        for (name, description, index) in custom_unit_column_infos:
            nwb_file.add_unit_column(name = name, description = description, index = index)
        for analysis_roi in self.all_analysis_rois:
            info = self._get_info_for_nwb_units_table_from_analysis_roi(analysis_roi)
            nwb_file.add_unit(**info)
        return nwb_file


    def _write_nwb_file(self, nwb_file: NWBFile) -> None:
        filepath = self.results_dir_path.joinpath('autogenerated_nwb_file.nwb')
        with NWBHDF5IO(filepath, 'w') as io:
            io.write(nwb_file)


    def _get_info_for_nwb_units_table_from_analysis_roi(self, analysis_roi) -> dict:
        info = {
            'ROI_label_ID': analysis_roi.label_id, 
            'total_peaks_count': analysis_roi.peaks_count, 
            'variance_area': np.nan, 
            'peaks_frame_idxs': [], 
            'peaks_bit_value': [], 
            'peaks_amplitude': [], 
            'peaks_df_over_f': [], 
            'peaks_auc': [], 
            'peaks_classification': []
        }
        if hasattr(analysis_roi, 'variance_area'):
            info['variance_area'] = analysis_roi.variance_area
        if analysis_roi.peaks_count > 0:
            info['peaks_frame_idxs'] = [peak.frame_idx for peak in analysis_roi.peaks.values()]
            info['peaks_bit_value'] = [peak.intensity for peak in analysis_roi.peaks.values()]
            if hasattr(analysis_roi, 'baseline'):
                info['peaks_amplitude'] = [peak.amplitude for peak in analysis_roi.peaks.values()]
                info['peaks_df_over_f'] = [peak.delta_f_over_f for peak in analysis_roi.peaks.values()]
                info['peaks_auc'] = [peak.area_under_curve for peak in analysis_roi.peaks.values()]
                info['peaks_classification'] = [peak.peak_type for peak in analysis_roi.peaks.values()]
        return info


    def _create_variance_area_dataframe(self, filtered_rois: list[AnalysisROI]) -> pd.DataFrame:
        data = {'ROI label ID': [],
                'Variance Area': []}
        for roi in filtered_rois:
            data['ROI label ID'].append(roi.label_id)
            data['Variance Area'].append(roi.variance_area)
        df = pd.DataFrame(data = data)
        return df
    
    
    def _create_and_save_csv_result_files(self, filtered_rois: list[AnalysisROI]) -> None:
        if len(filtered_rois) > 0:
            df_variance_areas = self._create_variance_area_dataframe(filtered_rois)
            df_variance_areas.to_csv(self.results_dir_path.joinpath('Variance_area_results.csv'), index = False)
            peak_results_per_roi = [results.export_peak_results_df_from_analysis_roi(roi) for roi in filtered_rois]
            df_all_peak_results = pd.concat(peak_results_per_roi, ignore_index = True)
            max_peak_count_across_all_rois = df_all_peak_results.groupby('ROI label ID').count()['peak frame index'].max()
            zfill_factor = int(np.log10(max_peak_count_across_all_rois)) + 1
            amplitude_and_delta_f_over_f_results_all_rois = []
            auc_results_all_rois = []
            for roi_label_id in df_all_peak_results['ROI label ID'].unique():
                tmp_df_single_roi = df_all_peak_results[df_all_peak_results['ROI label ID'] == roi_label_id].copy()
                amplitude_and_delta_f_over_f_results_all_rois.append(results.create_single_roi_amplitude_and_delta_f_over_f_results(tmp_df_single_roi, zfill_factor))
                auc_results_all_rois.append(results.create_single_roi_auc_results(tmp_df_single_roi, zfill_factor))
            df_all_amplitude_and_delta_f_over_f_results = pd.concat(amplitude_and_delta_f_over_f_results_all_rois, ignore_index = True)
            df_all_auc_results = pd.concat(auc_results_all_rois, ignore_index = True)
            # Once all DataFrames are created successfully, write them to disk 
            df_all_peak_results.to_csv(self.results_dir_path.joinpath('all_peak_results.csv'), index = False)
            df_all_amplitude_and_delta_f_over_f_results.to_csv(self.results_dir_path.joinpath('Amplitude_and_dF_over_F_results.csv'), index = False)
            df_all_auc_results.to_csv(self.results_dir_path.joinpath('AUC_results.csv'), index = False)

    
    def _create_and_save_individual_traces_pdf_result_file(self, filtered_rois: list[AnalysisROI]) -> None:
            filepath = self.results_dir_path.joinpath('Individual_traces_with_identified_events.pdf')
            with PdfPages(filepath) as pdf:
                for indicate_activity in [True, False]:
                    overview_fig, ax = results.plot_activity_overview(filtered_rois, self.recording.preview, indicate_activity, self.focus_area, self.grid_configs)
                    pdf.savefig(overview_fig)
                    plt.close()
                label_ids_overview_fig, ax = results.plot_rois_with_label_id_overview(filtered_rois, self.recording.preview, self.focus_area, self.grid_configs)
                pdf.savefig(label_ids_overview_fig)
                plt.close()
                for roi in filtered_rois:
                    fig = results.plot_intensity_trace_with_identified_peaks_for_individual_roi(roi)
                    pdf.savefig(fig)
                    plt.close()


    def _create_and_save_single_trace_results_as_csv(self, filtered_rois: list[AnalysisROI]) -> None:
        single_trace_subdir_path = self.results_dir_path.joinpath('single_traces')
        single_trace_subdir_path.mkdir(parents = True, exist_ok = True)
        for analysis_roi in filtered_rois:
            df_single_trace = self._initialize_single_trace_df(analysis_roi)
            for peak_idx in df_single_trace[df_single_trace['is_peak'] == True].index:
                peak = analysis_roi.peaks[peak_idx]
                df_single_trace.at[peak_idx, 'peak_type'] = peak.peak_type
                df_single_trace.at[peak_idx, 'amplitude'] = peak.amplitude
                df_single_trace.at[peak_idx, 'delta_f_over_f'] = peak.delta_f_over_f
                df_single_trace.at[peak_idx, 'has_baseline_intersections'] = peak.has_neighboring_intersections
                if peak.has_neighboring_intersections == True:
                    df_single_trace.at[peak_idx, 'pre_peak_intersection_idx'] = peak.frame_idxs_of_neighboring_intersections[0]
                    df_single_trace.at[peak_idx, 'post_peak_intersection_idx'] = peak.frame_idxs_of_neighboring_intersections[1]
                    df_single_trace.at[peak_idx, 'area_under_curve'] = peak.area_under_curve
            self._save_single_trace_df(df_single_trace, analysis_roi, single_trace_subdir_path)
            

    def _initialize_single_trace_df(self, analysis_roi: AnalysisROI) -> pd.DataFrame:
        is_peak_mask = np.zeros(analysis_roi.mean_intensity_over_time.shape[0], dtype=bool)
        is_peak_mask[analysis_roi.frame_idxs_of_peaks] = True
        data = {'intensity': analysis_roi.mean_intensity_over_time,
                'estimated_baseline': analysis_roi.baseline,
                'is_peak': is_peak_mask}
        df = pd.DataFrame(data = data)
        df.index.name = 'frame_idx'
        new_columns = ['peak_type', 
                       'amplitude', 
                       'delta_f_over_f', 
                       'has_baseline_intersections', 
                       'pre_peak_intersection_idx', 
                       'post_peak_intersection_idx', 
                       'area_under_curve']
        df[new_columns] = np.nan
        df = df.astype({'peak_type': 'object',
                        'has_baseline_intersections': 'object'})
        return df


    def _save_single_trace_df(self, 
                              df_single_trace: pd.DataFrame, 
                              analysis_roi: AnalysisROI,
                              single_trace_subdir_path: Path) -> None:
        filepath_friendly_roi_label_id = analysis_roi.label_id.replace('/', '-')
        filepath = single_trace_subdir_path.joinpath(f'data_of_ROI_{filepath_friendly_roi_label_id}.csv')
        df_single_trace.to_csv(filepath)



"""Defines data and processing structures."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_model.ipynb.

# %% auto 0
__all__ = ['Logger', 'Model']

# %% ../nbs/02_model.ipynb 10
# External functional dependencies:
from pathlib import Path
from datetime import datetime, timezone
from matplotlib.pyplot import show
import ipywidgets as w
import multiprocessing
from fastcore.basics import patch
import gc

# External dependencies for type hints:
from typing import Callable
from matplotlib.figure import Figure
from matplotlib.axes._axes import Axes

# Internal dependencies
from .datamodels import Config
from .processing import AnalysisJob
from .input import RecordingLoaderFactory, ROILoaderFactory, RecordingLoader, ROILoader, get_filepaths_with_supported_extension_in_dirpath, FocusAreaPathRestrictions


# %% ../nbs/02_model.ipynb 13
class Logger:

    """
    A simple logging utility class that captures log messages with UTC timestamps,
    allows retrieval and clearing of logs, and supports saving them to a file.
    """
    
    def __init__(self):
        self.logs = []

    def add_new_log(self, message: str) -> None:
        """
        Add a new log message with a UTC timestamp prefix. 
        The timestamp is formatted as 'dd-mm-yy HH:MM:SS.ffffff (UTC)'.
        """
        time_prefix_in_utc = datetime.now(timezone.utc).strftime('%d-%m-%y %H:%M:%S.%f')
        self.logs.append(f'{time_prefix_in_utc} (UTC): {message}')
        print(f'{time_prefix_in_utc} (UTC): {message}')

    def get_logs(self) -> list[str]:
        return self.logs

    def clear_logs(self) -> None:
        self.logs = []

    def save_current_logs(self, save_dir: Path) -> None:
        filepath = save_dir.joinpath('logs.txt')
        with open(filepath , 'w+') as logs_file:
            for log_message in self.logs:
                logs_file.write(f'{log_message}\n')
        print(f'Logs saved to {filepath}')


# %% ../nbs/02_model.ipynb 28
class Model:

    def __init__(self, 
                 config: Config | str
                ) -> None:
        self.num_processes = multiprocessing.cpu_count()
        self.analysis_job_queue = []
        self.result_directories = []
        self.logs = Logger()
        if isinstance(config, str):
            config = Config(data_source_path=config)
        self.config = config
        self.nwb_metadata = None
        self.gui_enabled = False
        self.callback_view_update_infos = None
        self.callback_view_show_output_screen = None
        self.view_output = None
        self.pixel_conversion = None

# %% ../nbs/02_model.ipynb 31
@patch
def create_analysis_jobs(self: Model
                        ) -> None:
    self._ensure_data_from_previous_jobs_was_removed()
    self.add_info_to_logs('Basic configurations for data import validated. Starting creation of analysis job(s)...', True)
    if self.config.batch_mode:
        all_subdir_paths_with_rec_file = self._get_all_subdir_paths_with_rec_file(self.config.data_source_path)
        all_subdir_paths_with_rec_file.sort()
        for idx, subdir_path in enumerate(all_subdir_paths_with_rec_file):
            if self.config.results_filepath:
                result_path = self.config.results_filepath / subdir_path.name
            else:
                result_path = subdir_path
            self._create_analysis_jobs_for_single_rec(subdir_path, result_path)
    else:
        self._create_analysis_jobs_for_single_rec()
    self.add_info_to_logs('All job creation(s) completed.', True, 100.0)

# %% ../nbs/02_model.ipynb 33
@patch
def _ensure_data_from_previous_jobs_was_removed(self: Model
                                               ) -> None:
    self.add_info_to_logs('Loading of new source data. All previously created jobs & logs will be deleted.', True)
    self.analysis_job_queue = []
    self.logs.clear_logs()


@patch
def _get_all_subdir_paths_with_rec_file(self: Model, 
                                        top_level_dir_path: Path
                                       ) -> list[Path]:
    rec_loader_factory = RecordingLoaderFactory()
    supported_extensions_for_recordings = rec_loader_factory.all_supported_extensions
    all_subdir_paths_that_contain_a_supported_recording_file = []
    for elem in top_level_dir_path.iterdir():
        if not elem.name.startswith('.'):
            if elem.is_dir():
                supported_recording_filepaths = [elem_2 for elem_2 in elem.iterdir() if elem_2.suffix in supported_extensions_for_recordings]
                if len(supported_recording_filepaths) > 0:
                    all_subdir_paths_that_contain_a_supported_recording_file.append(elem)
    return all_subdir_paths_that_contain_a_supported_recording_file



@patch
def _create_analysis_jobs_for_single_rec(self: Model, 
                                         recording_path: Path = None,
                                         result_path: Path = None
                                        ) -> None:
    if recording_path is None:
        if self.config.recording_filepath:
            recording_path = self.config.recording_filepath
        else:
            recording_path = self.config.data_source_path
    self.add_info_to_logs(f'Starting with Job creation(s) for {str(recording_path)}', True)
    recording_loader = self._get_recording_loader(recording_path)


    if self.config.roi_filepath:
        roi_filepath = self.config.roi_filepath
    else:
        roi_filepath = recording_loader.filepath.parent

    if self.config.roi_mode == 'file':
        roi_loaders = self._get_all_roi_loaders(roi_filepath)
    else:
        roi_loaders = None

    if self.config.focus_area_filepath:
        focus_area_filepath = self.config.focus_area_filepath
    else:
        focus_area_filepath = recording_loader.filepath.parent

    if result_path is None:
        if self.config.results_filepath:
            result_path = self.config.results_filepath
        else:
            result_path = recording_loader.filepath.parent

    if self.config.focus_area_enabled:
        focus_area_dir_path = self._get_focus_area_dir_path(focus_area_filepath)
        if focus_area_dir_path is None:
            analysis_job = self._create_single_analysis_job(recording_loader, roi_loaders, result_filepath=result_path)
            self.analysis_job_queue.append(analysis_job)
            self.add_info_to_logs(f'Successfully created a single job for {focus_area_filepath} at queue position: #{len(self.analysis_job_queue)}.', True)
        else:
            all_focus_area_loaders = self._get_all_roi_loaders(focus_area_dir_path)
            assert len(all_focus_area_loaders) > 0, f'Focus Area analysis enabled, but no focus area ROIs could be found. Please revisit your source data and retry!'
            for idx, focus_area_loader in enumerate(all_focus_area_loaders):
                if result_path is None:
                    if self.config.results_filepath:
                        result_path = self.config.results_filepath / focus_area_loader.filepath.stem
                analysis_job_with_focus_area = self._create_single_analysis_job(recording_loader, roi_loaders, focus_area_loader, result_filepath=result_path)
                self.analysis_job_queue.append(analysis_job_with_focus_area)
                job_creation_message = (f'Successfully created {idx + 1} out of {len(all_focus_area_loaders)} job(s) for {recording_path} '
                                        f'at queue position: #{len(self.analysis_job_queue)}.')
                self.add_info_to_logs(job_creation_message, True)
    else:
        analysis_job = self._create_single_analysis_job(recording_loader, roi_loaders, result_filepath=result_path)
        self.analysis_job_queue.append(analysis_job)
        self.add_info_to_logs(f'Successfully created a single job for {recording_path} at queue position: #{len(self.analysis_job_queue)}.', True)
    self.add_info_to_logs(f'Finished Job creation(s) for {recording_path}!', True)


@patch
def _get_recording_loader(self: Model, 
                          source_path: Path
                         ) -> RecordingLoader:
    rec_loader_factory = RecordingLoaderFactory()
    if source_path.is_dir():
        self.add_info_to_logs(f'Looking for a valid recording file in {source_path}...', True)
        valid_filepaths = get_filepaths_with_supported_extension_in_dirpath(source_path, rec_loader_factory.all_supported_extensions, 1)
        if len(valid_filepaths) == 0:
            self.add_info_to_logs(f'Could not find any recording files of supported type at {source_path}!', True)
        elif len(valid_filepaths) >  1:
            filepath = valid_filepaths[0]
            too_many_files_message = (f'Found more than a single recording file of supported type at {source_path}, i.e.: {valid_filepaths}. '
                                      f'However, only a single file was expected. NA3 continues with {filepath} and will ignore the other files.')
            self.add_info_to_logs(too_many_files_message, True)
        else:
            filepath = valid_filepaths[0]
            self.add_info_to_logs(f'Found recording file of supported type at: {filepath}.', True)
    else:
        filepath = source_path
        self.add_info_to_logs(f'Found recording file of supported type at: {filepath}.', True)
    recording_loader = rec_loader_factory.get_loader(filepath)
    return recording_loader


@patch
def _get_all_roi_loaders(self: Model, 
                         data_source_path: Path
                        ) -> list[ROILoader]:
    assert data_source_path.is_dir(), f'You must provide a directory as source data when using ROI mode or enabling Focus Areas. Please revisit your input data and retry.'
    roi_loader_factory = ROILoaderFactory()
    all_filepaths_with_supported_filetype_extensions = get_filepaths_with_supported_extension_in_dirpath(data_source_path, roi_loader_factory.all_supported_extensions)
    all_roi_loaders = [roi_loader_factory.get_loader(roi_filepath) for roi_filepath in all_filepaths_with_supported_filetype_extensions]
    return all_roi_loaders


@patch
def _get_focus_area_dir_path(self: Model, 
                             source_path: Path
                            ) -> Path:
    focus_area_path_restrictions = FocusAreaPathRestrictions()
    supported_dir_names = focus_area_path_restrictions.supported_dir_names
    if source_path.is_dir():
        source_dir_path = source_path
    else:
        source_dir_path = source_path.parent
    dirs_with_valid_name = [elem for elem in source_dir_path.iterdir() if (elem.name in supported_dir_names) & (elem.is_dir() == True)]
    if len(dirs_with_valid_name) == 0:
        no_dir_found_message = (f'You enabled Focus Area but a correspondingly named directory could not be found in {source_dir_path}. '
                                f'Please use one of the following for the name of the directory that contains the Focus Area ROIs: {supported_dir_names}. '
                                'In absence of such a directory, analysis is continued without using the Focus Area mode for this data.')
        self.add_info_to_logs(no_dir_found_message, True)
        focus_area_dir_path = None
    elif len(dirs_with_valid_name) > 1:
        too_many_dirs = (f'More than a single Focus Area directory was found in the following parent directory: {source_dir_path}, i.e.: '
                         f'{dirs_with_valid_name}. However, only the use of a single one that contains all your Focus Area ROIS is '
                         f'currently supported. {dirs_with_valid_name[0]} will be used for this analysis, while the other(s): {dirs_with_valid_name[1:]} '
                         'will be ignored to continue processing.')
        self.add_info_to_logs(too_many_dirs, True)
        focus_area_dir_path = dirs_with_valid_name[0]
    else:
        focus_area_dir_path = dirs_with_valid_name[0]
    return focus_area_dir_path


@patch
def _create_single_analysis_job(self: Model,
                                recording_loader: RecordingLoader,
                                roi_loaders: list[ROILoader] | None,
                                focus_area_loader: ROILoader = None,
                                result_filepath: Path = None
                               ) -> AnalysisJob:
    data_loaders = {'recording': recording_loader}
    if roi_loaders is not None:
        data_loaders['rois'] = roi_loaders
    if focus_area_loader is not None:
        data_loaders['focus_area'] = focus_area_loader
    return AnalysisJob(self.num_processes, data_loaders, result_filepath)

# %% ../nbs/02_model.ipynb 35
@patch
def run_analysis(self: Model
                ) -> None:
    self._display_configs()
    self.add_info_to_logs('Starting analysis...', True)
    for job_idx in range(len(self.analysis_job_queue)):
        analysis_job = self.analysis_job_queue.pop(0)
        self.add_info_to_logs(f'Starting to process analysis job with index #{job_idx}.')
        analysis_job.run_analysis(self.config)
        self.add_info_to_logs(f'Analysis successfully completed. Continue with creation of results.. ')
        analysis_job.create_results(self.config, self.nwb_metadata)
        self.add_info_to_logs(f'Results successfully created at: {analysis_job.results_dir_path}')
        if self.gui_enabled:
            self.callback_view_show_output_screen()
            with self.view_output:
                activity_overview_fig = analysis_job.activity_overview_plot[0]
                activity_overview_fig.set_figheight(400 * self.pixel_conversion)
                activity_overview_fig.tight_layout()
                show(activity_overview_fig)
        self._save_user_settings_as_json(analysis_job)
        self.result_directories.append(analysis_job.results_dir_path)
        self.add_info_to_logs('Updating all log files to contain all logs as final step. All valid logs files will end with this message.')
        self.logs.save_current_logs(analysis_job.results_dir_path)
    else:
        gc.collect()

# %% ../nbs/02_model.ipynb 38
@patch
def _display_configs(self: Model
                    ) -> None:
    self.add_info_to_logs('Configurations for Analysis Settings and Result Creation validated successfully.', True)
    self.add_info_to_logs(f'Analysis Settings are:')
    for line in self.config.display_all_attributes():
        self.add_info_to_logs(line)


@patch
def _save_user_settings_as_json(self: Model, 
                                analysis_job: AnalysisJob
                               ) -> None:
    filepath = analysis_job.results_dir_path.joinpath('user_settings.json')
    self.config.recording_filepath = analysis_job.recording.filepath
    if analysis_job.focus_area_enabled:
        self.config.focus_area_filepath = analysis_job.focus_area.filepath
    else:
        self.config.focus_area_filepath = None
    if analysis_job.rois_source == 'file':
        self.config.roi_filepath = [roi.filepath for roi in analysis_job.all_rois]
    with open(filepath, 'w+') as user_settings_json:
        user_settings_json.write(self.config.to_json())

# %% ../nbs/02_model.ipynb 40
@patch
def setup_connection_to_update_infos_in_view(self: Model, 
                                             update_infos: Callable
                                            ) -> None:
    """ Allows to configure the widget in the GUI that is used to display the logs. """
    self.callback_view_update_infos = update_infos
    self.gui_enabled = True 

# %% ../nbs/02_model.ipynb 42
@patch
def setup_connection_to_display_results(self: Model, 
                                        show_output_screen: Callable, 
                                        output: w.Output, 
                                        pixel_conversion: float
                                       ) -> None:
    """ Allows to configure the widget in the GUI that is used to display images, plots, and figures. """
    self.callback_view_show_output_screen = show_output_screen
    self.view_output = output
    self.pixel_conversion = pixel_conversion
    self.gui_enabled = True

# %% ../nbs/02_model.ipynb 45
@patch
def preview_window_size(self: Model, 
                        grid_size
                       ) -> tuple[Figure, Axes]:
    job_for_preview = self.analysis_job_queue[0]
    preview_fig, preview_ax = job_for_preview.preview_window_size(grid_size)
    return preview_fig, preview_ax

# %% ../nbs/02_model.ipynb 48
@patch
def add_info_to_logs(self: Model, 
                     message: str, 
                     display_in_gui: bool = False, 
                     progress_in_percent: float | None = None
                    ) -> None:
    self.logs.add_new_log(message)
    if (display_in_gui == True) and (self.gui_enabled == True): 
        self.callback_view_update_infos(message, progress_in_percent)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/08_api.ipynb.

# %% auto 0
__all__ = ['run_analysis']

# %% ../nbs/08_api.ipynb 2
from .datamodels import Config
from .model import Model

def run_analysis(config: Config | str) -> None:
    """
    Run analysis.

    Args:
        config (Config | str): Configuration for analysis or data source path to run analysis with default settings.
    """
    model = Model(config)
    model.create_analysis_jobs()
    model.run_analysis()
